/*
 * Copyright (c) 2015-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <vzctl/libvzctl.h>

#include "vzparsecpu.h"

#define CGROUP_ROOT	"/sys/fs/cgroup/"
#define CGROUP_VE	CGROUP_ROOT "ve"
#define CGROUP_CPU	CGROUP_ROOT "cpu"
#define CGROUP_MEM	CGROUP_ROOT "memory"

static int is_task_running(unsigned long pid)
{
	int res = 0;
	char path[PATH_MAX], state;
	FILE *fp;

	snprintf(path, sizeof(path), "/proc/%lu/stat", pid);
	fp = fopen(path, "r");
	if (!fp)
		return res;
	if (fscanf(fp, "%*u %*s %c", &state) == 1 && state == 'R')
		res = 1;
	fclose(fp);
	return res;
}

static int count_running_tasks(const char *entry)
{
	int count = 0;
	char path[PATH_MAX];
	FILE *fp;

	snprintf(path, sizeof(path), CGROUP_VE "/%s/cgroup.procs", entry);
	fp = fopen(path, "r");
	if (!fp)
		return count;
	while (!feof(fp)) {
		unsigned long pid;
		if (fscanf(fp, "%lu\n", &pid) != 1)
			break;
		if (is_task_running(pid) == 1)
			++count;
	}
	fclose(fp);
	return count;
}

static ssize_t read_text_file(const char *path, char *buf, size_t size)
{
	ssize_t count, total = 0, max = size - 1;
	int fd = open(path, O_RDONLY);
	if (fd == -1)
		return -1;
	do {
		count = read(fd, buf + total, max - total);
		if (count <= 0)
			break;
		total += count;
	} while (total < max);
	close(fd);
	if (count == -1)
		return -1;
	if (total > 0) {
		if (buf[total - 1] == '\n')
			buf[total - 1]= '\0';
		else
			buf[total] = '\0';
	}
	return total;
}

static int get_veid(const char *entry, unsigned *veid)
{
	ssize_t count;
	unsigned long veid_;
	char path[PATH_MAX], buf[128];

	snprintf(path, sizeof(path), CGROUP_VE "/%s/ve.veid", entry);
	count = read_text_file(path, buf, sizeof(buf));
	if (count == -1)
		return -1;

	if (sscanf(buf, "%lu", &veid_) != 1)
		return -1;
	*veid = veid_;
	return 0;
}

static int get_ve_weight(const char *entry)
{
	ssize_t count;
	int shares;
	char path[PATH_MAX], buf[21] = {};

	snprintf(path, sizeof(path), CGROUP_CPU "/%s/cpu.shares", entry);
	count = read_text_file(path, buf, sizeof(buf));
	if (count == -1)
		return 0;
	if (sscanf(buf, "%d", &shares) != 1)
		return 0;
	return shares <= 0 ? 0 : 512000 / shares;
}

static void add_ve0()
{
	if (!is_ve_in_set(0))
		return;
	s_line_t r = {};
	r.ve_uuid[0] = '0';
	r.running = count_running_tasks("");
	r.weight = get_ve_weight("");
	ve[1][numve[1]++] = r;
}

int cg_find_ves()
{
	int i, n;
	struct vzctl_ids *ctids = vzctl2_alloc_env_ids();
	if (!ctids)
		return -1;
	n = vzctl2_get_env_ids_by_state(ctids, ENV_STATUS_RUNNING);
	if (n < 0) {
		vzctl2_free_env_ids(ctids);
		return -1;
	}

	for (i = 0; i < n; ++i) {
		unsigned veid;
		const char *ctid = ctids->ids[i];
		if (get_veid(ctid, &veid))
			continue;
		if (is_ve_in_set(veid)) {
			s_line_t r = {};
			r.veid = veid;
			strncpy(r.ve_uuid, ctid, sizeof(r.ve_uuid) - 1);
			r.ve_uuid[sizeof(r.ve_uuid) - 1] = '\0';
			r.running = count_running_tasks(ctid);
			r.weight = get_ve_weight(ctid);
			ve[1][numve[1]++] = r;
		}
	}
	vzctl2_free_env_ids(ctids);

	if (show_ve0_mode)
		add_ve0();
	return 0;
}
