/*
 * Copyright (c) 2003-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/file.h>
#include <uuid/uuid.h>
#include <sys/types.h>
#include <dirent.h>

#include "vzparsecpu.h"

#define PRLCTL_BINARY "/usr/sbin/prlctl"

static int prl_get_vm_name(const char *uuid, char *buf, int size)
{
	char cmd[512];
	char pb[BUFSIZ];
	FILE *fp = NULL;
	int ret = -1;

	if (access(PRLCTL_BINARY, F_OK))
		return ret;

	snprintf(cmd, sizeof(cmd), "%s list -Ho name %s 2>/dev/null", PRLCTL_BINARY, uuid);
	fp = popen(cmd, "r");
	if (fp == NULL)
		return ret;

	while (fgets(pb, sizeof(pb), fp) != NULL) {
		snprintf(buf, size, "%s", pb);
		ret = 0;
	}

	pclose(fp);
	return ret;
}

#define UUID_MAP_FILE "/etc/parallels/uuid.map"

struct uuidmap_t {
	uuid_t uuid;
	unsigned int id;
};
static struct uuidmap_t *_uuidmap;
static int _uuidmap_size;

static int read_uuid_map()
{
	int fd, ret;
	struct stat st;

	if (_uuidmap != NULL) {
		free(_uuidmap);
		_uuidmap = NULL;
	}
	fd = open(UUID_MAP_FILE, O_RDONLY);
	if (fd == -1) {
		if (errno == ENOENT)
			return 0;
		return -1;
	}
	if (flock(fd, LOCK_SH))
		goto err;
	if (fstat(fd, &st))
		goto err;

	_uuidmap = malloc(st.st_size);
	if (_uuidmap == NULL)
		goto err;

	ret = read(fd, _uuidmap, st.st_size);
	if (ret == -1)
		goto err;

	_uuidmap_size = st.st_size / sizeof(struct uuidmap_t);

	flock(fd, LOCK_UN);
	close(fd);

	return 0;

err:
	if (_uuidmap != NULL) {
		free(_uuidmap);
		_uuidmap = NULL;
	}
	flock(fd, LOCK_UN);
	close(fd);

	return -1;
}

static struct uuidmap_t *find_uuid(unsigned int id)
{
	int i;

	if (_uuidmap != NULL) {
		for (i = 0; i < _uuidmap_size; i++) {
			if (_uuidmap[i].id == id)
				return &_uuidmap[i];
		}
	}
	return NULL;
}

void update_prl_vm_pid()
{
	DIR *proc;
	FILE *fp;
	s_line_t *vestat;
	int idx = 0;
	struct dirent *pde;
	char buf[1024];
	char proc_uuid[64];
	char *s;

	if (_uuidmap == NULL)
		return;

	proc = opendir("/proc");
	if(!proc)
		return;

	while ((pde = readdir(proc)) != NULL) {
		if (!isdigit(pde->d_name[0]))
			continue;

		sprintf(buf, "/proc/%s/cmdline", pde->d_name);
		fp = fopen(buf, "r");
		if (!fp)
			continue;
		if (fgets(buf, sizeof(buf), fp) != NULL &&
			(s = strstr(buf, "prl_vm_app")) != NULL)
		{
			strncpy(proc_uuid + 1, s + 12, 37);
			proc_uuid[0] = '{'; proc_uuid[37] = '}'; proc_uuid[38] = 0;
			while ((vestat = vzstatcpu_get_ve_stats(idx++, 1)) != NULL) {
				if (vestat->uuid != NULL && !strcmp(vestat->uuid, proc_uuid))
					vestat->prl_vm_pid = atoi(pde->d_name);
			}
		}

		fclose(fp);
	}
	closedir(proc);
}

void parse_uuid()
{
	s_line_t *vestat;
	s_line_t *prev_vestat;
	int idx = 0;

	read_uuid_map();

	while ((vestat = vzstatcpu_get_ve_stats(idx++, 1)) != NULL) {
		struct uuidmap_t *uuid;
		char id[64];
		char name[1024];

		if (vestat->uuid != NULL)
			continue;
		// FIXME: surpress stopped Vm
		if (vestat->numproc.held == 0)
			continue;
		uuid = find_uuid(vestat->veid);
		if (uuid != NULL) {
			vestat->env_type = VM_TYPE;
			uuid_unparse(uuid->uuid, id + 1);
			id[0] = '{'; id[37] = '}'; id[38] = 0;
			if (vestat->name == NULL) {
				// Check name already filled in previous measurement
				prev_vestat = vzstatcpu_get_ve_stats(idx - 1, 0);
				if (prev_vestat != NULL && prev_vestat->name != NULL)
					vestat->name = strdup(prev_vestat->name);
				else
					if (prl_get_vm_name(id, name, sizeof(name)) == 0)
						vestat->name = strdup(name);
			}
		} else {
			vestat->env_type = CT_TYPE;
			sprintf(id, "%lu", vestat->veid);
		}
		if (vestat->uuid != NULL)
			free(vestat->uuid);
		vestat->uuid = strdup(id);
	}

	update_prl_vm_pid();
}


