/*
 * Copyright (c) 2003-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/vzctl_venet.h>
#include <linux/vzctl_netstat.h>

#include "tc.h"
#include "vzparsecpu.h"

int get_ipaddr(char *name, unsigned int *ip)
{
	struct in_addr addr;

	if (name == NULL)
		return ERR_BAD_IP;
	if (!inet_aton(name, &addr))
		return ERR_BAD_IP;
	*ip = addr.s_addr;
	return 0;
}

char *get_ipname(unsigned int ip)
{
	struct in_addr addr;

	addr.s_addr = ip;
	return inet_ntoa(addr);
}

int tc_set_classes(struct vz_tc_class_info *c_info, int length)
{
	struct vzctl_tc_classes classes;

	classes.info = c_info;
	classes.length = length;
	if (ioctl(get_vzctlfd(), VZCTL_TC_SET_CLASS_TABLE, &classes))
	{
		fprintf(stderr, "Unable to set classes: %s\n", strerror(errno));
		return ERR_IOCTL;
	}
	return 0;
}

int tc_get_classes(struct vz_tc_class_info *c_info, int length)
{
	struct vzctl_tc_classes classes;
	int ret;

	classes.info = c_info;
	classes.length = length;
	if ((ret = ioctl(get_vzctlfd(), VZCTL_TC_GET_CLASS_TABLE, &classes)) < 0)
	{
		fprintf(stderr, "Unable to get classes: %s\n", strerror(errno));
		return ERR_IOCTL;
	}
	return ret;
}

inline int tc_get_ve_stats(struct vzctl_tc_get_stat *stat)
{
	int ret;

	if ((ret = ioctl(get_vzctlfd(), VZCTL_TC_GET_STAT, stat)) < 0)
	{
		if (errno != ESRCH)
		{
			fprintf(stderr, "Unable to get Container %d statistics: %s\n",
				stat->veid, strerror(errno));
			return ERR_IOCTL;
		}
		ret = 0;
	}
	stat->length = ret;
	return ret;
}

int tc_destroy_ve_stats(int veid)
{
	if (ioctl(get_vzctlfd(), VZCTL_TC_DESTROY_STAT, veid))
	{
		if (errno != ESRCH)
		{
			fprintf(stderr, "Reset failed: %s\n", strerror(errno));
			return ERR_IOCTL;
		}
	}
	printf("Counters for Container %d reseted.\n", veid);
	return 0;
}

int tc_destroy_all_stats()
{
	if (ioctl(get_vzctlfd(), VZCTL_TC_DESTROY_ALL_STAT))
	{
		fprintf(stderr, "Reset failed: %s\n", strerror(errno));
		return ERR_IOCTL;
	}
	printf("Counters reseted.\n");
	return 0;
}

int tc_get_class_num()
{
	int len;

	if ((len = ioctl(get_vzctlfd(), VZCTL_TC_CLASS_NUM, NULL)) < 0)
	{
		fprintf(stderr, "Unable get classes count: %s\n",
				strerror(errno));
		return ERR_IOCTL;
	}
	return len;
}

inline int tc_max_class()
{
	int len;

	if ((len = ioctl(get_vzctlfd(), VZCTL_TC_MAX_CLASS, NULL)) < 0)
	{
		fprintf(stderr, "Unable get max classes: %s\n", strerror(errno));
		return ERR_IOCTL;
	}
	return len;
}

int tc_get_stat_num()
{
	int len;

	if ((len = ioctl(get_vzctlfd(), VZCTL_TC_STAT_NUM, NULL)) < 0)
	{
		fprintf(stderr, "Unable get Container count in accounting table: %s\n",
			strerror(errno));
		return ERR_IOCTL;
	}
	return len;
}

int tc_get_ve_list(struct vzctl_tc_get_stat_list *velist)
{
	int ret;

	if ((ret = ioctl(get_vzctlfd(), VZCTL_TC_GET_STAT_LIST, velist)) < 0)
	{
		fprintf(stderr, "Unable to get Container list: %s\n",
				strerror(errno));
		return ERR_IOCTL;
	}
	velist->length = ret;
	return ret;
}

inline int get_ve_stats(int veid,  struct vzctl_tc_get_stat *stat)
{
	int len;

	if ((len = tc_max_class()) < 0)
		return len;
	if (!len)
		return 0;
	stat->incoming = calloc(len, sizeof(__u64));
	if (stat->incoming == NULL)
		return ERR_NOMEM;
	stat->outgoing = calloc(len, sizeof(__u64));
	if (stat->outgoing == NULL)
		return ERR_NOMEM;
	stat->incoming_pkt = calloc(len, sizeof(__u32));
	if (stat->incoming_pkt == NULL)
		return ERR_NOMEM;
	stat->outgoing_pkt = calloc(len, sizeof(__u32));
	if (stat->outgoing_pkt == NULL)
		return ERR_NOMEM;
	stat->veid = veid;
	stat->length = len;	
	if ((len = tc_get_ve_stats(stat)) < 0)
		return len;
	return 0;
}

int get_ve_list(struct vzctl_tc_get_stat_list *velist)
{
	int len;

	if ((len = tc_get_stat_num()) < 0)
		return len;
	if (!len )
		return 0;
	velist->list = calloc(len, sizeof(int));
	velist->length = len;
	if ((len = tc_get_ve_list(velist)) < 0)
		return len;
	return len;
}

int parse_int(const char *str, int *val)
{
        char *tail;

        if (!str || !val)
                return 1;
        errno = 0;
        *val = (int)strtol(str, (char **)&tail, 10);
        if (*tail != '\0' || errno == ERANGE)
                return 1;

        return 0;
}
