/*
 * Copyright (c) 2003-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/sysinfo.h>

#include "vzparsecpu.h"

#define MAXLINE 1024
#define CLAMP(x, maxv) ((x) > (maxv) ? (maxv) : (x))
#define max(x, y) ((x) > (y) ? (x) : (y))

#define SHOW_RUNONLY	1
#define SHOW_TOTAL	2

#define IGNORE_OVERFLOW 0         // 1 - to drop overflowed counters in used_time

s_line_t ve[3][MAXN];             /* 0,1 - last two measurements, 2-diff from ve[1]-ve[0] */
ve_avg_t ve_avg[MAXN];            /* average stats for VEs */
int numve[3],ve_avg_num;          /* num of VEs */
int HZ = 100;

char snapshot_date[2][MAXLINE];   /* dates of data */
char start_date[MAXLINE], last_date[MAXLINE];

/* totals per first-last snapshot */
unsigned long long total_time, total_cycles;
unsigned long total_cpuunits, total_cpuunitsr, total_tasks, total_running;
unsigned int nr_cpus;

/* kernel output versions */
int vestat_version;
int ub_version;

/* modes */
int gnuplot = 0;
int show_top5 = 0;
int show_top10 = 1;
int show_total_only = 1;

/* VE set when -v is given */
int ve_set_num = 0;
int ve_set[VE_SET_MAX];

/* average stats */
unsigned long long avg_total_time = 0, avg_cycles = 0;
unsigned avg_running = 0, avg_tasks = 0, avg_overflow = 0;
int stat_num = 0;

int show_ve0_mode = 0;

/* check if VE should be processed  */ 
int is_ve_in_set(int veid)
{
	int i;

	/* no set is given at all */
	if (ve_set_num == 0) return 1;

	for (i=0; i<ve_set_num; i++)
		if (ve_set[i] == veid) return 1;
	return 0;
}

int vzstatcpu_parse_vestat(FILE *f)
{
	char line[MAXLINE];

	// required for bsearch
	qsort(ve[1], numve[1], sizeof(s_line_t), ve_uuid_sort_fn);
	while (!feof(f))
	{
		if (!fgets(line, sizeof(line),f)) break;;
		if (feof(f)) line[0] = 0;
		if (!strcmp(line, "end /proc/vz/vestat\n")) break;
		if (parse_vestat_version(line)) continue;
		parse_vestat_line(line);
	}
	return 0;
}

int vzstatcpu_parse_ub(FILE *f)
{
	char line[MAXLINE];

	// required for bsearch
	qsort(ve[1], numve[1], sizeof(s_line_t), ve_uuid_sort_fn);
	while (!feof(f))
	{
		if (!fgets(line, sizeof(line),f)) break;
		if (feof(f)) line[0] = 0;
		if (!strcmp(line, "end /proc/user_beancounters\n")) break;
		if (parse_ub_version(line)) continue;
		parse_ub_line(line);
	}
	return 0;
}

int vzstatcpu_parse_veinfo(FILE *f)
{
	char line[MAXLINE * 16]; /* for 1024 ip addresses */

	// required for bsearch
	qsort(ve[1], numve[1], sizeof(s_line_t), ve_uuid_sort_fn);
	while (!feof(f))
	{
		if (!fgets(line, sizeof(line),f)) break;
		if (feof(f)) line[0] = 0;
		parse_veinfo_line(line);
	}
	return 0;
}

void parse_cgroup_cpu(s_line_t *vestat, const char *name, unsigned long *val)
{
	FILE *fp;
	char f[256];

	snprintf(f, sizeof(f), "/sys/fs/cgroup/cpuacct/machine.slice/%s/%s",
			vestat->ve_uuid, name);
	fp = fopen(f, "rt");
	if (fp == NULL)
		return;

	if (fscanf(fp, "%lu", val) != 1)
		*val = 0;
	fclose(fp);
}

void vzstatcpu_parse_cpu()
{
	s_line_t *vestat;
	int idx = 0;

	while ((vestat = vzstatcpu_get_ve_stats(idx++, 1)))
	{
		parse_cgroup_cpu(vestat, "cpu.shares", &vestat->weight);
		parse_cgroup_cpu(vestat, "cpu.rate", &vestat->rate);
	}
}

/* save date of snapshot */
int parse_date(char *line, int n)
{
	int res,d,m,y,h,min,s;
	static int calln = 0;

	res = sscanf(line, "%u/%u/%u %u:%u:%u\n", &m, &d, &y, &h, &min, &s);
	if (res != 6) return 0;
	snprintf(snapshot_date[n], MAXLINE, "%u/%02u/%02u %u:%02u:%02u", m, d, y, h, min, s);

	/* 1st data chunk */
	if (calln++ == 0)
		strcpy(start_date, snapshot_date[n]);
	return 1;
}

int parse_nr_cpus(char *line)
{
	int res, nr;

	res = sscanf(line, "nr_cpus=%d\n", &nr);
	if (res != 1) return 0;

	nr_cpus = nr;
	return 1;
}

int parse_vestat_version(char *line)
{
	int res, major, minor;
	static int calln = 0;

	/* can't change fairsched version on fly */
	if (calln != 0)
		return 0;
	res = sscanf(line, "Version: %d.%d\n", &major, &minor);
	if (res == 2)
	{
		calln++;
		vestat_version = MKVERSION(major, minor);
		return 1;
	}
	return 0;
}

int parse_vestat_line(char *line)
{
	s_line_t *r;
	int res;
	unsigned long user, nice, system, uptime, numsched, tmp;
	char ve_uuid[37] = {};
	cycles_t c_idle, c_strv, c_uptime, c_usedtime, c_maxlat, c_totlat;

	if (vestat_version == MKVERSION(2,1))
	{
		res = sscanf(line, "%36s %lu %lu %lu %lu "
			"%Lu %Lu %Lu %lu",
			ve_uuid, &user, &nice, &system, &uptime,
			&c_idle, &c_strv, &c_uptime, &tmp);
		/* not matching */
		if (res != 8)
			return 0;
	}
	else if (vestat_version == MKVERSION(2,2))
	{
		res = sscanf(line, "%36s %lu %lu %lu %lu "
			"%Lu %Lu %Lu %Lu "
			"%Lu %Lu %lu %lu",
			ve_uuid, &user, &nice, &system, &uptime,
			&c_idle, &c_strv, &c_uptime, &c_usedtime,
			&c_maxlat, &c_totlat, &numsched, &tmp);
		/* not matching */
		if (res != 12)
			return 0;
	}
	else
	{
		printf("Unknown /proc/vz/vestat version\n");
		exit(1);
	}

	/* find VE desc in list by veid */
	r = (s_line_t*) bsearch(ve_uuid, ve[1], numve[1], sizeof(s_line_t), ve_uuid_search_fn);
	if (!r)
		return 0;

	if (vestat_version == MKVERSION(2,2))
	{
		r->cycles = c_uptime >> 12;
		r->used_time = c_usedtime >> 12;
		r->maxlat = c_maxlat;
		r->totlat = c_totlat;
		r->numsched = numsched;
	}
	else
	{
		r->cycles = uptime;
		r->used_time = user+nice+system;
	}
	
	return 1;
}

int parse_ub_version(char *line)
{
	int res, major, minor;
	static int calln = 0;

	/* can't change fairsched version on fly */
	if (calln != 0)
		return 0;
	res = sscanf(line, "Version: %d.%d\n", &major, &minor);
	if (res == 2)
	{
		calln++;
		ub_version = MKVERSION(major, minor);
		return 1;
	}
	return 0;
}

int parse_ub_line(char *line)
{
	s_line_t *r;
	static char last_ve_uuid[37];
	ub_line_t ub;
	char ve_uuid[37] = {};
	char name[32], *s;
	int res;

	/* find and cut CTID first */
	s = strchr(line, ':');
	if (s)
	{
		char *p = line;
		while (p && isspace(*p))
			p++;
		res = sscanf(p, "%36[^:]", ve_uuid);
		if (res != 1) return 0;
		s++;
		strncpy(last_ve_uuid, ve_uuid, sizeof(last_ve_uuid) - 1);
		last_ve_uuid[sizeof(last_ve_uuid) - 1] = '\0';
	}
	else
	{
		s = line;
		strncpy(ve_uuid, last_ve_uuid, sizeof(ve_uuid) - 1);
		ve_uuid[sizeof(ve_uuid) - 1] = '\0';
	}

	/* now parse everything else */
	if (ub_version == MKVERSION(2,5))
	{
		res = sscanf(s, "%31s %lu %lu %lu %lu %lu",
			name, &ub.held, &ub.maxheld, &ub.barrier, &ub.limit, &ub.failcnt);
		/* not matching */
		if (res != 6) return 0;
	}
	else
	{
		printf("Unknown /proc/user_beancounters version\n");
		exit(1);
	}

	/* find VE desc in list by veid */
	r = (s_line_t*) bsearch(ve_uuid, ve[1], numve[1], sizeof(s_line_t), ve_uuid_search_fn);
	if (!r) return 0;

	r->fcnt += ub.failcnt;
	if (!strcmp(name, "kmemsize"))
		r->kmemsize = ub;
	else if (!strcmp(name, "privvmpages"))
		r->privvmpages = ub;
	else if (!strcmp(name, "numtcpsock"))
		r->numtcpsock = ub;
	else if (!strcmp(name, "numothersock"))
		r->numothersock = ub;
	else if (!strcmp(name, "numproc"))
		r->numproc = ub;

	return 1;
}

int parse_veinfo_line(char *line)
{
	s_line_t *r;
	int res;
	unsigned long class_id, nr_procs;
	char ve_uuid[37] = {};
	char ip_addr[32], *s;

	res = sscanf(line, "%36s %lu %lu %31s",
		ve_uuid, &class_id, &nr_procs, ip_addr);
	/* not matching */
	if (res < 3)
		return 0;
	if (res == 3)
		ip_addr[0] = 0;
	else {
		/* find the last IP in the list, since list is reversed */
		s = strrchr(line, ' ');
		res = 0;
		if (s != NULL)
			res = sscanf(s + 1, "%31s", ip_addr);
		if (res == 0)
			ip_addr[0] = 0;
	}

	/* find VE desc in list by veid */
	r = (s_line_t*) bsearch(ve_uuid, ve[1], numve[1], sizeof(s_line_t), ve_uuid_search_fn);
	if (!r) return 0;

	strncpy(r->ip_addr, ip_addr, sizeof(r->ip_addr));
	r->nr_procs = nr_procs;
	r->tasks = nr_procs;

	return 1;
}

/* search functions */
int ve_uuid_search_fn(const void* x1, const void* x2)
{
	return strcmp(x1, ((s_line_t*)x2)->ve_uuid);
}

/* sort functions */
int w_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->weight > ((s_line_t*)x2)->weight;
}

int id_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->veid > ((s_line_t*)x2)->veid;
}

int ve_uuid_sort_fn(const void* x1, const void* x2)
{
	return strcmp(x1, ((s_line_t*)x2)->ve_uuid);
}

int vmem_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->privvmpages.held <
	       ((s_line_t*)x2)->privvmpages.held;
}

int kmem_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->kmemsize.held < ((s_line_t*)x2)->kmemsize.held;
}

int sock_sort_fn(const void* x1, const void* x2)
{
	s_line_t *v1, *v2;

	v1 = (s_line_t*)x1;
	v2 = (s_line_t*)x2;
	return v1->numtcpsock.held + v1->numothersock.held <
	       v2->numtcpsock.held + v2->numothersock.held;
}

int mlat_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->maxlat < ((s_line_t*)x2)->maxlat;
}

int fcnt_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->fcnt < ((s_line_t*)x2)->fcnt;
}

int nrproc_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->tasks < ((s_line_t*)x2)->tasks;
}

int status_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->status < ((s_line_t*)x2)->status;
}

int usedtime_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->used_time < ((s_line_t*)x2)->used_time;
}

int nrrunning_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->running < ((s_line_t*)x2)->running;
}

int ttr_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->ttr < ((s_line_t*)x2)->ttr;
}

int avg_id_sort_fn(const void* x1, const void* x2)
{
	return ((ve_avg_t*)x1)->veid > ((ve_avg_t*)x2)->veid;
}

int avg_ut_sort_fn(const void* x1, const void* x2)
{
	return ((ve_avg_t*)x1)->used_time < ((ve_avg_t*)x2)->used_time;
}

int avg_w_sort_fn(const void* x1, const void* x2)
{
	return ((ve_avg_t*)x1)->cpuunits < ((ve_avg_t*)x2)->cpuunits;
}

int ior_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io_rate[IO_READ] <
			((s_line_t*)x2)->io_rate[IO_READ];
}

int iow_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io_rate[IO_WRITE] <
			((s_line_t*)x2)->io_rate[IO_WRITE];
}

int iort_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io[IO_READ] <
			((s_line_t*)x2)->io[IO_READ];
}

int iowt_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io[IO_WRITE] <
			((s_line_t*)x2)->io[IO_WRITE];
}

int iod_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io_rate[IO_DIRTY] <
			((s_line_t*)x2)->io_rate[IO_DIRTY];
}

int ios_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->io_rate[IO_SYNC] <
			((s_line_t*)x2)->io_rate[IO_SYNC];
}

int cpu_iowait_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->cpu_rate[CPU_IOWAIT] <
			((s_line_t*)x2)->cpu_rate[CPU_IOWAIT];
}

int cpu_iowait_avg_sort_fn(const void* x1, const void* x2)
{
	return ((s_line_t*)x1)->cpu_rate_avg[CPU_IOWAIT] <
			((s_line_t*)x2)->cpu_rate_avg[CPU_IOWAIT];
}

#define DEF_SORT_FUNC(name, field)				\
	int sort_fn_##name (const void* x1, const void* x2)	\
	{							\
		return ((s_line_t*)x1)->field < 		\
			((s_line_t*)x2)->field;			\
	}							\
	void vzstatcpu_sort_##name(void)			\
	{							\
		qsort(ve[2], numve[2], sizeof(s_line_t),	\
			sort_fn_##name);			\
	}

DEF_SORT_FUNC(rx,   net.rate[NET_RX]);
DEF_SORT_FUNC(tx,   net.rate[NET_TX]);
DEF_SORT_FUNC(rxt,  net.total[NET_RX]);
DEF_SORT_FUNC(txt,  net.total[NET_TX]);
DEF_SORT_FUNC(rxp,  net.rate[NET_RX_PKTS]);
DEF_SORT_FUNC(txp,  net.rate[NET_TX_PKTS]);
DEF_SORT_FUNC(rxpt, net.total[NET_RX_PKTS]);
DEF_SORT_FUNC(txpt, net.total[NET_TX_PKTS]);

DEF_SORT_FUNC(ioused, io_rate[IO_USED_TM]);
DEF_SORT_FUNC(iowait, io_rate[IO_WAIT_TM]);
DEF_SORT_FUNC(ioidle, io_rate[IO_IDLE_TM]);
DEF_SORT_FUNC(iospeed, io_rate[IO_TRANSFERRED_SEC]);

void show_w_stats(int n)
{
	unsigned int w=0, m=0;
	int i;
	FILE *f = stdout;

	qsort(ve[n], numve[n], sizeof(s_line_t), w_sort_fn);

	if (gnuplot)
		f=fopen("ve_weight.data", "wt");
	else
		printf("cpuunits-CTs: ");
	if (!f) {perror("Couldn't open data file: ve_weight.data"); exit(1);}
	for (i=0; i<numve[n]; i++)
	{
		if (ve[n][i].weight != w)
		{
			if (w !=0 && m != 0) fprintf(f, gnuplot ? " %u\n" : "%u ", m);
			w = ve[n][i].weight; m=0;
			fprintf(f, gnuplot ? "%d " : "%d-", w);
		}

//    if (!gnuplot) fprintf(f, "%u ", ve[n][i].veid);
		m++;
	}
	if (w !=0 && m != 0) fprintf(f, gnuplot ? " %u\n" : "%u", m);
	if (gnuplot) fclose(f); else printf("\n");
}

/* print msg padded with '-' on left and right */
void show_barred(const char* msg)
{
	char bar[80];
	int len;

	len = strlen(msg);
	if (len < 80)
	{
		len = 80-len;
		memset(bar, '-', sizeof(bar));
		bar[(len+1)/2] = 0;
		printf("%s\n", bar);
		printf("%s\n", msg);
		bar[len/2] = 0;
		printf("%s\n", bar);
	}
	else
		printf("%s\n", msg);
}

void show_top(s_line_t *ve, int count, int n, char *subtitle, int opt)
{
	int i, overflow=0;

	if (n == 0) return;

	printf("%7s %17s %22s %16s\n", subtitle, "usage", "config", "tasks");
	printf("%7s %9s %8s %8s  %8s %8s %7s %8s\n",
		"CTID", "cycles", "cpuunits", "percents", "cpuunits", "percents", "total", "running");
	for (i=0; i<n && i<count; i++) {
		unsigned long cpuunits;

		if ((opt & SHOW_RUNONLY) && !ve[i].running)
			continue;
		cpuunits = (!ve[i].cycles) ? 0 :
			(unsigned long)((unsigned long long)ve[i].used_time
				* total_cpuunits / ve[i].cycles / nr_cpus);
		printf("%7lu %9lu %8lu %8.1f%1s %8lu %8.1f %7lu %8lu \n",
			ve[i].veid,
			ve[i].used_time,
			cpuunits,
			ve[i].cpu_prcntg,
			ve[i].overflow ? "?" : "",
			ve[i].cpuunits,
			ve[i].cpuunits_prcntg,
			ve[i].tasks, ve[i].running);
		if (ve[i].overflow) overflow = 1;
	}
	// there is not reason to show total for set of 1 VE
	if ((opt & SHOW_TOTAL) && (ve_set_num != 1))
	{
		printf("Total  %10Lu %8s %8.1f%1s %8lu %8s %7lu %8lu\n",
			total_time, "", (!total_cycles) ? 0 :
				100.0 * total_time / total_cycles / nr_cpus,
			overflow ? "?" : "", total_cpuunits, "",
			total_tasks, total_running);
	}
	printf("\n");
}

void vzstatcpu_sort_cpu(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), usedtime_sort_fn);
}

void vzstatcpu_sort_veid(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), id_sort_fn);
}

void vzstatcpu_sort_ve_uuid(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), ve_uuid_sort_fn);
}

void vzstatcpu_sort_vmem(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), vmem_sort_fn);
}

void vzstatcpu_sort_kmem(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), kmem_sort_fn);
}

void vzstatcpu_sort_sock(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), sock_sort_fn);
}

void vzstatcpu_sort_mlat(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), mlat_sort_fn);
}

void vzstatcpu_sort_fcnt(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), fcnt_sort_fn);
}

void vzstatcpu_sort_nrrunning(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), nrrunning_sort_fn);
}

void vzstatcpu_sort_nrproc(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), nrproc_sort_fn);
}

void vzstatcpu_sort_status(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), status_sort_fn);
}

void vzstatcpu_sort_ior(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), ior_sort_fn);
}

void vzstatcpu_sort_iow(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), iow_sort_fn);
}

void vzstatcpu_sort_iort(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), iort_sort_fn);
}

void vzstatcpu_sort_iowt(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), iowt_sort_fn);
}

void vzstatcpu_sort_iod(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), iod_sort_fn);
}

void vzstatcpu_sort_ios(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), ios_sort_fn);
}

void vzstatcpu_sort_cpu_iowait(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), cpu_iowait_sort_fn);
}

void vzstatcpu_sort_cpu_iowait_avg(void)
{
	qsort(ve[2], numve[2], sizeof(s_line_t), cpu_iowait_avg_sort_fn);
}

void show_utime_stats(int n)
{
	vzstatcpu_sort_cpu();
	show_top(ve[n], numve[n], show_top10 ? 10 : numve[n],
		show_top10 ? "Top 10": "", SHOW_TOTAL);
}

s_line_t* vzstatcpu_get_ve_stats(int idx, int n)
{
	if (idx >= numve[n])
		return NULL;
	return &ve[n][idx];
}

void show_utime_plot()
{
	int i;
	FILE *f;
/*  FILE *fplot; */
	char filename[MAXLINE];
	static int n = 0;

	if (!gnuplot) return;
	qsort(ve[2], numve[2], sizeof(s_line_t), usedtime_sort_fn);
/*
	fplot = fopen("ve_utime.plot", "at");
	if (!fplot) return;
*/
	for (i=0; i<numve[2]; i++)
	{
#if IGNORE_OVERFLOW
		if (ve[2][i].overflow) continue;
#endif

		snprintf(filename, MAXLINE, "ve_utime.%lu", ve[2][i].veid);
		f = fopen(filename, "at");
		if (!f) return;
		fprintf(f, "%s: ", snapshot_date[1]);
		fprintf(f, " %3.1f", (!ve[2][i].cycles) ? 0 :
			CLAMP((100.0 * ve[2][i].used_time) / ve[2][i].cycles
				/ nr_cpus, 100.0));
		fprintf(f, "\n");
		fclose(f);
/*    
		if (n == 0) fprintf(fplot, "plot ");
		if (n < numve[2]) fprintf(fplot, "\"%s\" using 1:3 with lines", filename);
		if (n < numve[2]) fprintf(fplot, ",");
*/
		n++;
	}
/*  if (n < numve[2]+1) fprintf(fplot, "\"ve_utime.sum\" using 1:3 with lines\n");
	fclose(fplot);
*/
	/* summary stats */
	f = fopen("ve_utime.sum", "at");
	if (!f) return;
	fprintf(f, "%s: ", snapshot_date[1]);
	fprintf(f, " %3.1f", (!total_cycles) ? 0 :
			CLAMP((100.0*total_time)/total_cycles/nr_cpus, 100.0));
	fprintf(f, "\n");
	fclose(f);
}

void show_avg_utime_plot()
{
#define AVG_TOP 5
	int i;
	char fn1[MAXLINE], fn2[MAXLINE];

	if (!gnuplot) return;
	qsort(ve_avg, ve_avg_num, sizeof(ve_avg_t), avg_ut_sort_fn);

	for (i=0; i<ve_avg_num && i<AVG_TOP; i++)
	{
		snprintf(fn1, MAXLINE, "ve_utime.%lu", ve_avg[i].veid);
		snprintf(fn2, MAXLINE, "ve_utime.%lu.top", ve_avg[i].veid);
		rename(fn1, fn2);
	}
}

void show_nr_stats(int n)
{
	qsort(ve[n], numve[n], sizeof(s_line_t), nrrunning_sort_fn);

	if (!show_top5 || !ve[n][0].running) return;

	show_top(ve[n], numve[n], 5, "NR 5", SHOW_RUNONLY);
}

void show_ttr_stats(int n)
{
	qsort(ve[n], numve[n], sizeof(s_line_t), ttr_sort_fn);

	if (!show_top5 || ve[n][0].ttr == 0) return;

	show_top(ve[n], numve[n], 5, "TTR 5", SHOW_RUNONLY);
}

void calc_stats(int n)
{
	int i;
	float w=0;

	total_time = 0; total_cycles = 0;
	total_cpuunits = total_cpuunitsr = total_tasks = total_running = 0;

	/* time to reply */
	for (i=0; i<numve[n]; i++) {
		if (ve[n][i].weight) {
			ve[n][i].cpuunits = ve[n][i].weight;
			if (ve[n][i].running) {
				/* add only active VEs weights */
				w += 1.0/ve[n][i].weight;
				total_cpuunitsr += ve[n][i].cpuunits;
				total_running += ve[n][i].running;
			}
			ve[n][i].ttr = w / (1.0 / ve[n][i].weight)
				* ve[n][i].running / HZ;
		}

		if (ve[n][i].env_type == VM_TYPE)
			ve[n][i].cpu_prcntg = CLAMP(ve[n][i].cpu_rate[CPU_USER] +
					ve[n][i].cpu_rate[CPU_SYSTEM] +
					ve[n][i].cpu_rate[CPU_NICE], 100.0);
		else if (ve[n][i].cycles)
			ve[n][i].cpu_prcntg = CLAMP((100.0*ve[n][i].used_time)/
					ve[n][i].cycles/nr_cpus, 100.0);
		else
			ve[n][i].cpu_prcntg = 0;

		total_tasks += ve[n][i].tasks;
		total_cpuunits += ve[n][i].cpuunits;
#if IGNORE_OVERFLOW
		if (!ve[n][i].overflow)
		{
			total_time += ve[n][i].used_time;
			total_cycles = max(total_cycles, ve[n][i].cycles);
		}
		else
			avg_overflow = 1;
#else
		total_time += ve[n][i].used_time;
		total_cycles = max(total_cycles, ve[n][i].cycles);
		if (ve[n][i].overflow)
			avg_overflow = 1;
#endif
	}

	// pass 2
	for (i=0; i<numve[n]; i++) {
		ve[n][i].cpuunits_prcntg = total_cpuunits ?
			(100.0*ve[n][i].cpuunits)/total_cpuunits : 0.0;
	}
}

void show_avg_w_stats()
{
	unsigned int w=0, m=0;
	int i;

	qsort(ve_avg, ve_avg_num, sizeof(ve_avg_t), avg_w_sort_fn);

	printf("cpuunits-CTs: ");
	for (i=0; i<ve_avg_num; i++)
	{
		if (ve_avg[i].cpuunits != w)
		{
			if (w !=0 && m != 0) printf("%u ", m);
			w = ve_avg[i].cpuunits; m=0;
			printf("%d-", w);
		}
		m++;
	}
	if (w !=0 && m != 0) printf("%u", m);
	printf("\n");
}

void show_avg()
{
	int i, ve_num;
	unsigned long avg_total_cpuunits = 0;

	if (stat_num < 1) return;

	if (!show_total_only) show_barred("");
	printf("CPU usage from %s to %s\n", start_date, last_date);
	show_avg_w_stats();

	qsort(ve_avg, ve_avg_num, sizeof(ve_avg_t), avg_ut_sort_fn);
	/* calc total cpuunits in our list */
	for (i=0; i<ve_avg_num; i++)
		avg_total_cpuunits += ve_avg[i].cpuunits;

	printf("%7s %20s %23s %16s\n", show_top10 ? "Top 10": "", "usage", "config", "tasks");
	printf("%7s %12s %8s %8s  %8s %8s %8s %8s\n",
		"CTID", "cycles", "cpuunits", "percents", "cpuunits", "percents", "total", "running");
	ve_num = show_top10 ? 10 : ve_avg_num;
	for (i=0; i<ve_num && i<ve_avg_num; i++)
	{
		unsigned long cpuunits = 0;
		float percents = 0;

		if (ve_avg[i].cycles) {
			cpuunits = (unsigned long long)ve_avg[i].used_time
				* avg_total_cpuunits / ve_avg[i].cycles
				/ nr_cpus;
			percents = CLAMP(100.0 * ve_avg[i].used_time
					/ ve_avg[i].cycles / nr_cpus, 100.0);
		}
		printf("%7lu %12Lu %8lu %8.1f%1s %8lu %8.1f %8.1f %8.1f\n",
			ve_avg[i].veid,
			ve_avg[i].used_time,
			cpuunits,
			percents,
			ve_avg[i].overflow ? "?" : "",
			ve_avg[i].cpuunits,
			(!avg_total_cpuunits) ?
				0.0 : (float)(100.0 * ve_avg[i].cpuunits
						/ avg_total_cpuunits),
			(float)ve_avg[i].tasks/(stat_num-1), (float)ve_avg[i].running/(stat_num-1));
	}

	printf("Total:  %12Lu %8s %8.1f%1s %8lu %8s %8.1f %8.1f\n",
		avg_total_time, "", (!avg_cycles) ? 0 :
			CLAMP(100.0*avg_total_time/avg_cycles/nr_cpus, 100.0),
		avg_overflow ? "?" : "",
		avg_total_cpuunits, "",
		(float)avg_tasks/(stat_num-1), (float)avg_running/(stat_num-1));
}

void make_diff()
{
	int i, j=0, n=0, k;

	qsort(ve[0], numve[0], sizeof(s_line_t), id_sort_fn);
	qsort(ve[1], numve[1], sizeof(s_line_t), id_sort_fn);

	for (i=0; i<numve[0]; i++)
	{
again:
		if (j >= numve[1]) break;
		if (ve[0][i].veid > ve[1][j].veid)
			{j++; goto again; }
		if (ve[0][i].veid < ve[1][j].veid)
			continue;

		ve[2][n] = ve[1][j];
		ve[2][n].uuid = NULL;
		ve[2][n].name = NULL;
		ve[2][n].hostname = NULL;

		/* no overflow when unsigned is used */
		ve[2][n].used_time -= ve[0][i].used_time;
		ve[2][n].cycles -= ve[0][i].cycles;
		if (ve[1][j].fcnt > ve[0][i].fcnt)
			ve[2][n].fcnt_delta = ve[1][j].fcnt - ve[0][i].fcnt;
		else
			ve[2][n].fcnt_delta = 0;
		ve[2][n].overflow = 0;
		if (ve[1][j].used_time < ve[0][i].used_time)
			ve[2][n].overflow = 1;
		/* if VE was down, then we can only use used_time after startup */
		if (ve[2][n].used_time > ve[2][n].cycles && ve[2][n].overflow)
			ve[2][n].used_time = ve[1][i].used_time;

		//calc cpu rate
		ve[2][n].cpu_sum = 0;
		for (k = 0; k < CPU_MAX; k++) {
			ve[2][n].cpu_sum += (ve[1][j].cpu[k] - ve[0][i].cpu[k]);
		}
		//prevent div to zero
		if (ve[2][n].cpu_sum == 0)
			ve[2][n].cpu_sum = 1;

		for (k = 0; k < CPU_MAX; k++) {
			ve[2][n].cpu_rate[k] =
					100.0 * (ve[1][j].cpu[k] - ve[0][i].cpu[k]) /
					(ve[2][n].cpu_sum);
		}

		//calc accumulative
		ve[1][j].cpu_stat_num = ve[0][i].cpu_stat_num + 1;
		for (k = 0; k < CPU_MAX; k++) {
			float prev = ve[0][i].cpu_rate_avg[k] * ve[0][i].cpu_stat_num;
			ve[2][n].cpu_rate_avg[k] =
				( prev + ve[2][n].cpu_rate[k] ) /
					ve[1][j].cpu_stat_num;

			//save for next cicle
			ve[1][j].cpu_rate_avg[k] = ve[2][n].cpu_rate_avg[k];
		}

		/* Update IO statistics */
		ve[2][n].io_msecs -= ve[0][i].io_msecs;
		
		for (k = 0; k < IO_MAX; k++) {
			if (!ve[2][n].io_msecs)
				ve[2][n].io_rate[k] = 0.;
			else
				ve[2][n].io_rate[k] =
					(ve[1][j].io[k] - ve[0][i].io[k]) /
					(ve[2][n].io_msecs/1000.0);
		}

		/* Update network statistics */
		ve[2][n].net.msecs -= ve[0][i].net.msecs;

		for (k = 0; k < NET_MAX; k++) {
			if (!ve[2][n].net.msecs)
				ve[2][n].net.rate[k] = 0.;
			else
				ve[2][n].net.rate[k] = (ve[1][j].net.total[k] - 
					ve[0][i].net.total[k]) /
						(ve[2][n].net.msecs/1000.0);
		}

		j++; n++;
	}
	numve[2] = n;
}

void make_ve_avg()
{
	int i, j=0, n=0;

	qsort(ve[2], numve[2], sizeof(s_line_t), id_sort_fn);
	qsort(ve_avg, ve_avg_num, sizeof(ve_avg_t), avg_id_sort_fn);

	for (i=0; i<numve[2]; i++)
	{
again:
		/* some VEs are beyond avg list max veid */
		if (j >= ve_avg_num) goto add_ve;

		/* ve_avg(j) is not present now, keep it in avg */
		if (ve[2][i].veid > ve_avg[j].veid)
			{j++; goto again;}
		/* ve(i) is not in avg list - add it */
		if (ve[2][i].veid < ve_avg[j].veid)
			goto add_ve;

		/* avg_ve[j].veid == ve[2][i].veid */
		if (!ve[2][i].overflow || !IGNORE_OVERFLOW)
		{
			ve_avg[j].cycles += ve[2][i].cycles;
			ve_avg[j].used_time += ve[2][i].used_time;
			ve_avg[j].tasks += ve[2][i].tasks;
			ve_avg[j].running += ve[2][i].running;
			if (ve[2][i].overflow)
				ve_avg[j].overflow = 1;
		}
		else
			ve_avg[j].overflow = 1;

		/* update cpuunits if changed */
		ve_avg[j].cpuunits = ve[2][i].cpuunits;
		j++;
		continue;

add_ve:
		/* VE is not in list yet - add it */
		ve_avg[ve_avg_num+n].veid = ve[2][i].veid;
		ve_avg[ve_avg_num+n].cycles = ve[2][i].cycles;
		ve_avg[ve_avg_num+n].used_time = ve[2][i].used_time;
		ve_avg[ve_avg_num+n].overflow = ve[2][i].overflow;
		ve_avg[ve_avg_num+n].tasks = ve[2][i].tasks;
		ve_avg[ve_avg_num+n].running = ve[2][i].running;
		ve_avg[ve_avg_num+n].cpuunits = ve[2][i].cpuunits;
		n++;
	}
	/* list can only grow */
	ve_avg_num += n;
}

void calc_avg()
{
	cycles_t max_cycles = 0;
	int i;

	for (i=0; i<numve[2]; i++)
		max_cycles = max(max_cycles, ve[2][i].cycles);
	avg_total_time += total_time;
	avg_cycles += max_cycles;
	avg_running += total_running;
	avg_tasks += total_tasks;
	stat_num++;
}

void vzstatcpu_update_stats()
{
	/* compare VEs state */
	make_diff();
	calc_stats(2);

	/* avg requires diff and stats */
	make_ve_avg();
	calc_avg();
}

static void free_sline(s_line_t *p)
{
	free(p->uuid);
	free(p->name);
	free(p->hostname);
}

void free_ves()
{
	int i, j;
	for (i = 0; i < 3; ++i)
		for (j = 0; j < numve[i]; ++j)
			free_sline(&ve[i][j]);
}

void vzstatcpu_begin_stats()
{
	int i;
	for (i = 0; i < numve[0]; ++i)
		free_sline(&ve[0][i]);

	memcpy(ve[0], ve[1], sizeof(ve[0]));

	for (i = 0; i < numve[1]; ++i) {
		s_line_t *l = &ve[1][i];
		l->uuid = NULL;
		l->name = NULL;
		l->hostname = NULL;
	}

	numve[0] = numve[1];
	numve[1] = 0;
	strcpy(snapshot_date[0], snapshot_date[1]);
	strcpy(last_date, snapshot_date[1]);
	snapshot_date[1][0] = 0;
}

void show_stats()
{
	if (numve[2] == 0) return;
	if (!show_total_only)
	{
		if (stat_num > 1) show_barred("");
		printf("Statistics for period from %s to %s\n", snapshot_date[0], snapshot_date[1]);
	}

	/* actually 2nd VE is latest so we use it as a pivot */
	if (!show_total_only)
	{
		show_w_stats(2);
		show_utime_stats(2);
		show_nr_stats(2);
		show_ttr_stats(2);
	}
	show_utime_plot();
}

int vzstatcpu_init()
{
	nr_cpus = get_nprocs_conf();
	numve[0] = numve[1] = 0;

	HZ = sysconf(_SC_CLK_TCK);
	if (!HZ)
		HZ = 100;

	return 0;
}
