/*
 * Copyright (c) 2003-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#ifndef __VZPARSECPU_H__
#define __VZPARSECPU_H__

#define VE_SET_MAX 10240

#define MKVERSION(x, y) (((x)<<16) + (y))
#define CLAMP(x, maxv) ((x) > (maxv) ? (maxv) : (x))
#define max(x, y) ((x) > (y) ? (x) : (y))

/* IO statistics fields */
enum {
	/* /proc/bc/VEID/ioacct */
	IO_READ = 0,
	IO_WRITE,
	IO_DIRTY,
	IO_CANCEL,
	IO_MISSED,
	IO_SYNC,

	/* /proc/bc/VEID/iostat */
	IO_USED_TM,
	IO_WAIT_TM,
	IO_IDLE_TM,
	IO_REQ_COMPLETED,
	IO_TRANSFERRED_SEC,

	/* /proc/vz/fairsched/VEID/cpu.delayacct.total */
	IO_SWAPIN_COUNT,

	IO_MAX
};

/* Network statistics fields */
enum {
	NET_RX = 0,
	NET_TX,
	NET_RX_PKTS,
	NET_TX_PKTS,
	NET_MAX,
};

/* CPU statistics fields */
enum {
	CPU_USER = 0,
	CPU_SYSTEM,
	CPU_NICE,
	CPU_IDLE,
	CPU_IOWAIT,

	CPU_MAX
};


typedef unsigned long long cycles_t;

typedef struct ub_line
{
	unsigned long held, maxheld, barrier, limit, failcnt;
} ub_line_t;

enum {
	CT_TYPE = 0,
	VM_TYPE,
};

struct slm_info_res {
	unsigned long long quality;
	unsigned long long inst_cap;
	unsigned long long avg_cap;
	unsigned long long inst_usage;
	unsigned long long avg_usage;
	unsigned long long limit;
};

struct slm_info {
	struct slm_info_res mem;
	struct slm_info_res lowmem;
	unsigned long long  oomlimit;
};

typedef struct s_line
{
	char ve_uuid[37];
	// /proc/fairsched
	unsigned long veid, cpus_real, cycles, id,
		parent, weight, rate, tasks, running, cpus, used_time;
	unsigned long cpuunits;
	unsigned long long start_tag, value, delay;
	unsigned long wakeup_max, ready;		// v2.6

	// /proc/vz/vestat
	cycles_t idle, uptime, maxlat, totlat;
	unsigned long numsched;

	// /proc/user_beancounters
	ub_line_t kmemsize, privvmpages, numproc, numtcpsock, numothersock;
	unsigned long fcnt, fcnt_delta;

	// /proc/bc/$CTID/meminfo
	unsigned long mem_total, mem_free;
	unsigned long swap_total, swap_free;

	// /proc/vz/veinfo
	unsigned long nr_procs;
	char ip_addr[32];
	char *uuid;
	int env_type;
	char *name;

	// /proc/stat
	unsigned long cpu[CPU_MAX];
	float cpu_rate[CPU_MAX];
	//average
	float cpu_rate_avg[CPU_MAX];
	//user + nice + system + cidle
	unsigned long long cpu_sum, cpu_stat_num;

	// calculated
	int overflow;
	int status, fcnt_st, maxlat_st;
	float ttr;
	float cpu_prcntg, cpuunits_prcntg;
	/* /etc/sysconfig/vz-scripts/veid.conf */
	char *hostname;

	/* I/O  statistics */
	unsigned long long io[IO_MAX];
	unsigned long long io_limit;
	unsigned long long iops_limit;

	/* Calculated from io rates */
	float io_rate[IO_MAX];

	unsigned long io_msecs;

	/* Network accounting */
	struct {
		unsigned long long      total[NET_MAX];
		float                   rate[NET_MAX];
		unsigned long           msecs;
	} net;

	/* SLM is on */
	int slm;
	struct slm_info slm_info;
	int prl_vm_pid;
} s_line_t;

typedef struct s_ve_avg
{
	unsigned long veid;
	unsigned long long used_time, cycles;
	unsigned long tasks, running;
	unsigned long cpuunits;
	int overflow;
} ve_avg_t;

extern unsigned long long total_time, total_cycles;
extern unsigned long total_cpuunits, total_cpuunitsr, total_tasks, total_running;

/* modes */
extern int gnuplot, show_top5, show_top10, show_total_only;
extern int ve_set_num, ve_set[];

/* prototypes */
int ve_uuid_search_fn(const void* x1, const void* x2);
int ve_uuid_sort_fn(const void* x1, const void* x2);
int id_sort_fn(const void* x1, const void* x2);

int parse_date(char *line, int n);
int parse_vestat_line(char *line);
int parse_vestat_version(char *line);
int parse_ub_version(char *line);
int parse_ub_line(char *line);
int parse_veinfo_line(char *line);
int parse_nr_cpus(char *line);
int parse_slm_info(unsigned long veid, struct slm_info *info);
void parse_uuid();

void show_stats();
void show_avg();
void show_avg_utime_plot();
void update_stats();

int vzstatcpu_init();
int vzstatcpu_parse_vestat(FILE *f);
int vzstatcpu_parse_ub(FILE *f);
int vzstatcpu_parse_veinfo(FILE *f);
void vzstatcpu_begin_stats();
void vzstatcpu_update_stats();
void vzstatcpu_parse_cpu();

void vzstatcpu_sort_cpu(void);
void vzstatcpu_sort_veid(void);
void vzstatcpu_sort_ve_uuid(void);
void vzstatcpu_sort_fcnt(void);
void vzstatcpu_sort_nrrunning(void);
void vzstatcpu_sort_nrproc(void);
void vzstatcpu_sort_status(void);
void vzstatcpu_sort_vmem(void);
void vzstatcpu_sort_kmem(void);
void vzstatcpu_sort_sock(void);
void vzstatcpu_sort_mlat(void);
void vzstatcpu_sort_iow(void);
void vzstatcpu_sort_ior(void);
void vzstatcpu_sort_iowt(void);
void vzstatcpu_sort_iort(void);
void vzstatcpu_sort_iod(void);
void vzstatcpu_sort_ios(void);
void vzstatcpu_sort_rx(void);
void vzstatcpu_sort_tx(void);
void vzstatcpu_sort_rxt(void);
void vzstatcpu_sort_txt(void);
void vzstatcpu_sort_rxp(void);
void vzstatcpu_sort_txp(void);
void vzstatcpu_sort_rxpt(void);
void vzstatcpu_sort_txpt(void);
void vzstatcpu_sort_ioused(void);
void vzstatcpu_sort_iowait(void);
void vzstatcpu_sort_ioidle(void);
void vzstatcpu_sort_iospeed(void);
void vzstatcpu_sort_cpu_iowait(void);
void vzstatcpu_sort_cpu_iowait_avg(void);

s_line_t* vzstatcpu_get_ve_stats(int idx, int n);
int get_vzctlfd();
int is_ve_in_set(int veid);

#define MAXN 10240
extern s_line_t ve[3][MAXN];
extern int numve[3];
extern int show_ve0_mode;

void free_ves();

#endif
