/*
 * Copyright (c) 2003-2017, Parallels International GmbH
 * Copyright (c) 2017-2019 Virtuozzo International GmbH. All rights reserved.
 *
 * This file is part of OpenVZ. OpenVZ is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include <sys/types.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <locale.h>
#include <langinfo.h>
#include <utmp.h>
#include <time.h>
#include <ncurses.h>
#include <sys/vfs.h>
#include <stdarg.h>
#include <ctype.h>
#include <term.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <termios.h>
#include <values.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/vzctl_venet.h>
#include <linux/vzctl_netstat.h>
#include <uuid/uuid.h>
#include "vziolimit.h"

#include "tc.h"
#include "vzparsecpu.h"
#include "cgroup.h"

enum {
	SCALE_NONE = 0,
	SCALE_BYTES,
	SCALE_KBYTES,
	SCALE_MBYTES,
	SCALE_GBYTES,
};

static int scale = SCALE_NONE;

typedef struct meminfo_struct
{
	unsigned long mem_total, low_total, high_total, swap_total;
	unsigned long mem_free, low_free, high_free, swap_free;
	unsigned long active, inactive, cached, buffers;
} meminfo_t;

typedef struct perfstat_struct
{
	unsigned long num;
	cycles_t max, total;
	cycles_t max_phys, total_phys;
} perfstat_t;

typedef struct lat_struct
{
	unsigned long num, prev_num;
	cycles_t max, time, prev_time;
	cycles_t avg[3];
} latstat_t;

#define PARSE_KSTAT_FUNC(str, param_name, struct_name)			\
do {									\
	sscanf(str, param_name " %lu %Lu %Lu %Lu %Lu",			\
		&struct_name.num,					\
		&struct_name.max_phys,					\
		&struct_name.max,					\
		&struct_name.total_phys,				\
		&struct_name.total);					\
} while (0)

typedef struct diskinfo_struct
{
	char *dev;
	char *mnt;
	int inuse;
	unsigned long long disk_total, inodes_total;
	unsigned long long disk_free, inodes_free;
} diskinfo_t;

typedef struct diskstat_struct
{
	unsigned int  major;
	unsigned int  minor;
	unsigned int  dr;
	unsigned long long  rblk;
	float delta_rblk;
	unsigned long long  wblk;
	float delta_wblk;
} diskstat_t;

typedef struct zone_struct
{
	char *name;
	char *fragmentation;
	unsigned long active_avg[3], inactive_avg[3], free_avg[3];
} zone_info_t;

typedef struct ifstats
{
	char *name;
	float rx, tx;
	unsigned long old_rx, old_tx;
	float pr, pt;
	unsigned long old_pr, old_pt;
} ifstat_t;

#define PROC_MEMINFO	"/proc/meminfo"
#define PROC_UPTIME	"/proc/uptime"
#define PROC_LOADAVG	"/proc/loadavg"
#define PROC_STAT	"/proc/stat"
#define PROC_VMSTAT	"/proc/vmstat"
#define PROC_VZ_STAT	"/sys/fs/cgroup/cpu/%s%s/cpu.proc.stat"
#define PROC_VZ_DELAY	"/sys/fs/cgroup/cpu/%s%s/cpu.delayacct.total"
#define PROC_SLABINFO	"/proc/slabinfo"
#define PROC_DISKSTATS	"/proc/diskstats"
#define PROC_NET_DEV	"/proc/net/dev"
#define PROC_MOUNTS	"/proc/mounts"

#define PROC_VZ_STATS	"/proc/vz/stats"
#define PROC_VZ_VESTAT	"/proc/vz/vestat"
#define PROC_VZ_VEINFO	"/proc/vz/veinfo"
#define PROC_VZ_LATENCY	"/proc/vz/latency"
#define PROC_VZ_MMPERF	"/proc/vz/mmperf"
#define PROC_UB		"/proc/user_beancounters"
#define PROC_VZ_IOACCT	"/proc/bc/%s/ioacct"
#define VE_CONF_DIR	"/etc/vz/conf/%s.conf"
#define PROC_VZ_IOSTAT	"/proc/bc/%s/iostat"
#define PROC_VZ_IOPRIO	"/proc/bc/%s/ioprio"
#define PROC_VZ_MEMINFO	"/proc/bc/%s/meminfo"

#define STR_MAX		"MAX"
#define STR_SLM_LIMIT	"-  "

#define KEY_ESCAPE	27
#define MB_SHIFT	20

/* Color ids	*/
#define C_NORMAL	1
#define C_YELLOW	2
#define C_RED		4
#define C_GREEN		8
#define C_INVERT	9
/* Status names */
#define	FAIL_ST		"[CRIT]"
#define OK_ST		"[ OK ]"

#define MAX_CHARS	512

static int pstat_mode = 0;
/*
 * Default screen dimensions
 */
static int screen_cols = MAX_CHARS;
static int avoid_last_column = 0;

/*
 * screen_rows will be obtained from terminal settings.
 * This default value will be used only if we can't get terminal
 * settings, for example when stdout is a file - not a terminal.
 */
static int screen_rows = 25;
static void set_win_size (void);
static volatile int need_resize;
static int term_saved;
static struct termios saved_tty;
static void screen_open(void);
static void screen_close(void);

static void exit_error (const char *fmt, ...)
{
	char ebuf[256];
	va_list ap;

	screen_close();

	if (fmt) {
		va_start(ap, fmt);
		vsnprintf(ebuf, sizeof(ebuf), fmt, ap);
		va_end(ap);

		fprintf(stderr, ebuf);
	}
	fflush(stdout);
	fflush(stderr);

	exit(-1);
}

typedef struct field_struct {
	char	*spec;		/* format specifier */
	char	*head;		/* default header in the POSIX locale */
	int	width;		/* field width (in characters) */
	char	*format;	/* output format (alignment) */
	char	sort_opt;	/* sort key */
	void	(*sort_fn)(void); /* sort function */
	char	*sort_desc;	/* sort option description */
	int	next_id;	/* next sort option for this field */
	int	used;		/* field is already specified */
} field_struct;

enum {
	ID_ID = 1,
	ID_ST, ID_VM, ID_SW, ID_KM, ID_PR, ID_PR2, ID_CPU, ID_SOCK,
	ID_FCNT, ID_MLAT, ID_IP, ID_UUID, ID_NAME, ID_HOST,
	ID_IOR, ID_IOW, ID_IORT, ID_IOWT, ID_IOD, ID_IOS,
	ID_RX, ID_TX, ID_RXT, ID_TXT, ID_RXP, ID_TXP, ID_RXPT, ID_TXPT,
	ID_IO_USED, ID_IO_WAIT, ID_IO_IDLE, ID_IO_AVG_SIZE,
	ID_IO, ID_IOPS, ID_SWAPIN,
	ID_CPU_IOWAIT,ID_CPU_IOWAIT_AVG,
};

/* Many of these are placeholders for unsupported options. */
#define SORT(x)        vzstatcpu_sort_##x
static field_struct field_ar[] = {

          /* spec    head     w  format sort sort_fn       sort_desc */
[ID_ID]   = {"id",   "CTID",  8, "%*s", 'n', SORT(veid),   "CTID" },
[ID_ST]   = {"st",   "ST",    2, "%*s", 't', SORT(status), "status" },
[ID_VM]   = {"vm",   "%%VM ", 7, "%*s", 'v', SORT(vmem),   "virtual memory" },
[ID_SW]   = {"sw",   "%%SW", 11, "%*s", 0,   NULL,         NULL },
[ID_KM]   = {"km",   "%%KM ", 7, "%*s", 'k', SORT(kmem),   "kernel memory" },
[ID_PR]   = {"proc", "PROC", 10, "%*s", 'r', SORT(nrrunning), "running procs",
								ID_PR2 },
/* Just a sort key */
[ID_PR2]  = { NULL,  NULL,    0, NULL,  'p', SORT(nrproc), "total procs" },
[ID_CPU]  = {"cpu",  "CPU ",  8, "%*s", 'c', SORT(cpu),    "CPU usage" },
[ID_SOCK] = {"sock", "SOCK",  7, "%*s", 's', SORT(sock),   "sock usage" },
[ID_FCNT] = {"fcnt", "FCNT",  4, "%*s", 'f', SORT(fcnt),   "fail counter" },
[ID_MLAT] = {"mlat", "MLAT",  4, "%*s", 'm', SORT(mlat),   "memory latency" },
[ID_IP]   = {"ip",   "IP",   14, "%-*s", 0,  NULL,         NULL },
[ID_UUID] = {"uuid", "ID",   39, "%-*s", 0,  NULL,         NULL },
[ID_NAME] = {"name", "NAME", 39, "%-*s", 0,  NULL,         NULL },
/* New features */
[ID_HOST] = {"host", "HOST",  8, "%-*s", 0,  NULL,         NULL },
[ID_IOR]  = {"ior",  "IOR",   8, "%*s", 'i', SORT(ior),    "IO read" },
[ID_IOW]  = {"iow",  "IOW",   8, "%*s", 'o', SORT(iow),    "IO write" },
[ID_IORT] = {"iort", "IORT", 11, "%*s", 'I', SORT(iort),   "IO read total" },
[ID_IOWT] = {"iowt", "IOWT", 11, "%*s", 'O', SORT(iowt),   "IO write total" },
[ID_IOD]  = {"iod",  "IOD",   8, "%*s", 'd', SORT(iod),    "IO dirty" },
[ID_IOS]  = {"ios",  "IOS",   8, "%*s", 'y', SORT(ios),    "IO sync" },
[ID_RX]   = {"rx",   "RX",    8, "%*s", '1', SORT(rx),     "Receive rate" },
[ID_TX]   = {"tx",   "TX",    8, "%*s", '2', SORT(tx),     "Transmit rate" },
[ID_RXT]  = {"rxt",  "RXT",   8, "%*s", '3', SORT(rxt),    "Receive total" },
[ID_TXT]  = {"txt",  "TXT",   8, "%*s", '4', SORT(txt),    "Transmit total" },
[ID_RXP]  = {"rxp",  "RXP",   6, "%*s", '5', SORT(rxp),    "Rcv packets rate" },
[ID_TXP]  = {"txp",  "TXP",   6, "%*s", '6', SORT(txp),    "Tx packets rate" },
[ID_RXPT] = {"rxpt", "RXPT",  8, "%*s", '7', SORT(rxpt),   "Rcv packets total"},
[ID_TXPT] = {"txpt", "TXPT",  8, "%*s", '8', SORT(txpt),   "Tx packets total" },

[ID_IO_USED] = {"ioused", "IOUSED", 6, "%*s", '9', SORT(ioused),	"IO used" },
[ID_IO_WAIT] = {"iowait", "IOWAIT", 6, "%*s", 'q', SORT(iowait),	"IO wait" },
[ID_IO_IDLE] = {"ioidle", "IOIDLE", 6, "%*s", 'w', SORT(ioidle),	"IO idle" },
[ID_IO_AVG_SIZE] = {"ioavg", "IOAVG",  8, "%*s", 0, NULL,	"IO avg request size" },

[ID_IO] =   {"io",   "IO",    11, "%*s", 0, NULL,	"Total IO rate vs limit" },
[ID_IOPS] = {"iops", "IOPS",  8, "%*s", 0, NULL,	"IOPS load vs limit" },

[ID_SWAPIN] = {"swapin", "SWAPIN",  8, "%*s", 	0, NULL,   "SWAPIN operation rate" },
[ID_CPU_IOWAIT] = {"cpu_iowait", "CPU IOWAIT",  8, "%*s", 'a', SORT(cpu_iowait),   "CPU IO wait" },
[ID_CPU_IOWAIT_AVG] = {"cpu_iowait_acc", "ACC IOWAIT",  8, "%*s", 'b', SORT(cpu_iowait_avg),   "Accumulated CPU IO wait" },
};

#define AR_SIZE(x)	(sizeof(x)/sizeof(x[0]))
#define FIELDS_COUNT	AR_SIZE(field_ar)
#define FIELDS_AR_SIZE	(FIELDS_COUNT * sizeof(int))

/* Default field set */
static int field_set_def[FIELDS_COUNT] =
	{ ID_ID, ID_ST, ID_VM, ID_KM, ID_PR, ID_CPU, ID_SOCK,
		ID_FCNT, ID_MLAT, ID_IP, };

static int field_set_pstat_def[FIELDS_COUNT] =
	{ ID_ST, ID_VM, ID_KM, ID_CPU,
		ID_FCNT, ID_MLAT, ID_NAME};

/* IO acct field set */
static int field_set_io[FIELDS_COUNT] =
	{ ID_ID, ID_ST, ID_IOR, ID_IOW, ID_IO, ID_IOPS, ID_IORT, ID_IOWT, ID_IP, };

static int field_set_pstat_io[FIELDS_COUNT] =
	{ ID_ST, ID_IOR, ID_IOW, ID_IO, ID_IOPS, ID_IORT, ID_IOWT, ID_NAME, };

/* IO stat field set */
static int field_set_io_stat[FIELDS_COUNT] =
	{ ID_ID, ID_ST, ID_IO_USED, ID_IO_WAIT, ID_IO, ID_IOPS, ID_IP, };

static int field_set_pstat_io_stat[FIELDS_COUNT] =
	{ ID_ST, ID_IO_USED, ID_IO_WAIT, ID_IO, ID_IOPS, ID_NAME, };


/* Netstat field set */
static int field_set_net[FIELDS_COUNT] =
	{ ID_ID, ID_ST, ID_RX, ID_TX, ID_RXT, ID_TXT,
	  ID_RXPT, ID_TXPT, ID_IP, };

static int field_set_pstat_net[FIELDS_COUNT] =
	{ ID_ST, ID_RX, ID_TX, ID_RXT, ID_TXT,
	  ID_RXPT, ID_TXPT, ID_NAME, };

/* Preloaded field set (for -O option) */
static int field_set_def_pre[]  = { ID_ID, ID_ST };
static int field_set_def_pre_pstat[] = { ID_ST };
static int field_set_def_post[] = { ID_IP };
static int field_set_def_post_pstat[] = { ID_NAME };

/* Current field set */
static int *field_set = field_set_def;

int has_proc_diskstats = 0;

unsigned long page_size = 4096;
int sort_mode = -1;
int idle_ve_display = 1;	// show idle VEs
int avg_mode = 0;
int hostname_mode = 0;
int name_mode = 1;
int io_mode = 0;
int io_stat_mode = 0;
int timeout_specified = 0;
extern int show_ve0_mode;
int net_mode = 0;
int clear_full_screen = 0;
char info_str[MAX_CHARS];

int has_color = -1;
int gen_text_output;
int gen_nofmt_output;

/*
 *  Fail criterias.
 *  Values overcommited this criterias will be collored.
 */
/* Load average */
int lavg_fail_level = 30;
/* Proc */
int proc_run_fail_level = 20;
int proc_unint_fail_level = 20;
/* sys & idle (%) */
int cpu_idle_fail_level = 10;
int cpu_sys_fail_level = 50;
/* CPU latency max/avg (ms) */
float cpu_latmax1_fail_level = 750;
float cpu_latmax2_fail_level = 1000;
float cpu_latavg1_fail_level = 400;
float cpu_latavg2_fail_level = 750;
/* Mem latency max/avg (ms) */
float mem_latmax1_fail_level = 500;
float mem_latmax2_fail_level = 1000;
float mem_latavg1_fail_level = 400;
float mem_latavg2_fail_level = 750;
/* Zones */
float zone_act1_fail_level = 8;
float zone_act2_fail_level = 4;
float zone_act_abs1_fail_level = 1;
float zone_act_abs2_fail_level = 1;
int zone_pages_fail_level = 100;
/* Swap (Kb) */
int swap_free1_fail_level = 75;
int swap_free2_fail_level = 50;
float swap_in1_fail_level = 0.5;
float swap_in2_fail_level = 1.0;
float swap_out1_fail_level = 0.5;
float swap_out2_fail_level = 1.0;
int swap_lat_avg1_fail_level = 500;
int swap_lat_avg2_fail_level = 750;
int swap_lat_max1_fail_level = 750;
int swap_lat_max2_fail_level = 1000;
/* Disk in (%) */
int disk_free_inodes1_fail_level = 20;
int disk_free_inodes2_fail_level = 5;
int disk_free_space1_fail_level = 20;
int disk_free_space2_fail_level = 5;
/*Disk load vs limit % */
int disk_load1_level = 50;
int disk_load2_level = 90;

/* VE delta fcnt */
int ve_delta_fcnt_fail_level = 1;

void print_status(int fail);
void printc(int color, char *fmt, ...);
void printe(char *fmt, ...);

meminfo_t meminfo;					// meminfo

#define MAX_DISK	20
diskinfo_t diskinfo[MAX_DISK];				// diskinfo
diskstat_t diskstat[MAX_DISK];

unsigned long uptime_secs, idle_secs;			// uptime
float loadavg[3];					// loadavg
unsigned long user_j, nice_j, system_j, idle_j, iowait_j; // stat
unsigned long long total_cpu_clk;
float swap_in, swap_out;
unsigned long slab_total, slab_active,			// slabinfo
	slab_inode, slab_dentry, slab_bh, slab_pb;
ifstat_t ifstats[10];					// /proc/net/dev
unsigned int ifcount = 0;

/* /proc/vz/latency */
int vz_latency_version;
latstat_t sched, a0, k0, u0, k1, u1, swapin;

/* /proc/vz/stats */
int vz_stats_version;
unsigned long venum, vmnum, unint_avg[3];
char proc_num_str[128], swap_cache_str[128];
unsigned long cycles_per_jiffy = 1;
unsigned long jiffies_per_second = 100;

#define CYCLES_PER_SEC ((unsigned long long)cycles_per_jiffy*jiffies_per_second)

zone_info_t zone_info[3];

/* /proc/vz/mmperf */
int vz_mmperf_version;
perfstat_t swapout, swapout_prev;

/* /proc/slabinfo */
int proc_slabinfo_version;

int sleep_time = 1000;
int loop_num = 0;
int num_disp = 0;

#define MAKE_MASK(x)		(1UL << x)
#define BIT_IS_SET(x, y)	(x & MAKE_MASK(y))

/* Display flags */
#define D_UPTIME_INFO	1
#define D_LOADAVG	2
#define D_PROC_INFO	3
#define D_CPU_INFO	4
#define D_MEM_INFO	5
#define D_MEM_ZONES	6
#define D_MEM_ZONES_F	7
#define D_MEM_LAT	8
#define D_SLAB_INFO	9
#define D_SWAP_INFO	10
#define D_SWAPC_INFO	11
#define D_SWAP_LAT	12
#define D_NET_INFO	13
#define D_NET_IFACE	14
#define D_DISK_INFO	15
#define D_DISK_MNT	16
#define D_NET_VETH	17

#define VL_BRIEF	(MAKE_MASK(D_UPTIME_INFO)|	\
			MAKE_MASK(D_LOADAVG)	|	\
			MAKE_MASK(D_PROC_INFO)	|	\
			MAKE_MASK(D_CPU_INFO)	|	\
			MAKE_MASK(D_MEM_INFO)	|	\
			MAKE_MASK(D_SWAP_INFO)	|	\
			MAKE_MASK(D_NET_INFO)	|	\
			MAKE_MASK(D_DISK_INFO))

#define VL_NORMAL	(VL_BRIEF 		|	\
			MAKE_MASK(D_MEM_ZONES)	|	\
			MAKE_MASK(D_MEM_LAT)	|	\
			MAKE_MASK(D_SLAB_INFO)	|	\
			MAKE_MASK(D_NET_IFACE))

#define VL_VERBOSE	(VL_NORMAL		|	\
			MAKE_MASK(D_MEM_ZONES_F)|	\
			MAKE_MASK(D_DISK_MNT)	|	\
			MAKE_MASK(D_SWAP_LAT)	|	\
			MAKE_MASK(D_SWAPC_INFO))

unsigned long disp_flags = VL_NORMAL;

typedef struct config_struct
{
	char *opt;
	void *param;
} config_t;
config_t config[] = {
	{"LOAD_AVG=%d",		&lavg_fail_level},
	{"PROC_RUN=%d",		&proc_run_fail_level},
	{"PROC_UNINT=%d",	&proc_unint_fail_level},
	{"CPU_IDLE=%d",		&cpu_idle_fail_level},
	{"CPU_SYS=%d",		&cpu_sys_fail_level},

	{"CPU_LAT_MAX_WARN=%f",	&cpu_latmax1_fail_level},
	{"CPU_LAT_MAX_ERR=%f",	&cpu_latmax2_fail_level},
	{"CPU_LAT_AVG_WARN=%f",	&cpu_latavg1_fail_level},
	{"CPU_LAT_AVG_ERR=%f",	&cpu_latavg2_fail_level},

	{"MEM_LAT_MAX_WARN=%f",	&mem_latmax1_fail_level},
	{"MEM_LAT_MAX_ERR=%f",	&mem_latmax2_fail_level},
	{"MEM_LAT_AVG_WARN=%f",	&mem_latavg1_fail_level},
	{"MEM_LAT_AVG_ERR=%f",	&mem_latavg2_fail_level},

	{"MEM_ZONE_ACT_INACT_FREE_WARN=%f",	&zone_act1_fail_level},
	{"MEM_ZONE_ACT_INACT_FREE_ERR=%f",	&zone_act2_fail_level},
	{"MEM_ZONE_ACT_INACT_FREE_ABS_WARN=%f",	&zone_act_abs1_fail_level},
	{"MEM_ZONE_ACT_INACT_FREE_ABS_ERR=%f",	&zone_act_abs2_fail_level},
	{"MEM_ZONE_ORDER_GT_0=%d",	&zone_pages_fail_level},

	{"SWAP_FREE_WARN=%d",	&swap_free1_fail_level},
	{"SWAP_FREE_ERR=%d",	&swap_free2_fail_level},
	{"SWAP_IN_WARN=%f",	&swap_in1_fail_level},
	{"SWAP_IN_ERR=%f",	&swap_in2_fail_level},
	{"SWAP_OUT_WARN=%f",	&swap_out1_fail_level},
	{"SWAP_OUT_ERR=%f",	&swap_out2_fail_level},
	{"SWAP_LAT_AVG_WARN=%d",&swap_lat_avg1_fail_level},
	{"SWAP_LAT_AVG_ERR=%d",	&swap_lat_avg2_fail_level},
	{"SWAP_LAT_MAX_WARN=%d",&swap_lat_max1_fail_level},
	{"SWAP_LAT_MAX_ERR=%d",	&swap_lat_max2_fail_level},

	{"DISK_FREE_INODES_WARN=%d",	&disk_free_inodes1_fail_level},
	{"DISK_FREE_INODES_ERR=%d",	&disk_free_inodes2_fail_level},
	{"DISK_FREE_SPACE_WARN=%d",	&disk_free_space1_fail_level},
	{"DISK_FREE_SPACE_ERR=%d",	&disk_free_space2_fail_level},

	{"DISK_IO_LOAD_WARN=%d",	&disk_load1_level},
	{"DISK_IO_LOAD_ERR=%d",	&disk_load2_level},


	/* VE_FAILCNT_DELTA left for backward compatibility */
	{"VE_FAILCNT_DELTA=%d",		&ve_delta_fcnt_fail_level},
	{"CT_FAILCNT_DELTA=%d",		&ve_delta_fcnt_fail_level},

	{NULL, NULL}
};


int *get_field_set(int *set)
{
	if (pstat_mode) {
		if (set == field_set_def)
			return field_set_pstat_def;
		else if (set == field_set_io)
			return field_set_pstat_io;
		else if (set == field_set_io_stat)
			return field_set_pstat_io_stat;
		else if (set == field_set_net)
			return field_set_pstat_net;
	}
	return set;
}
#define CONFIG	"/etc/vz/vzstat.conf"
#define FILTER_MAX_CTID 20
int veids_specified = 0;
char *veids[FILTER_MAX_CTID];
int default_fs = 1;		// by default display only default_fs_names[]
char *default_fs_names[] = {"ext2", "ext3", "ext4", "reiserfs", NULL};

#define FORMAT_STR_LEN	40	/* Maximum characters per line */
#define FORMAT_STR_EOL	" \\\n\t\t"
static void format_buf_add_eol(char **s, int *len)
{
	if (*len < FORMAT_STR_LEN)
		return;
	strcpy(*s, FORMAT_STR_EOL);
	*s += strlen(FORMAT_STR_EOL);
	*len = 0;
}

/*
 * Calculate length and allocate a buffer to hold the list of all possible
 * keys separated by commas.
 */
char *alloc_format_buf(void)
{
	unsigned f, len = 0;
	char *p;

	/* calculate string length */
	for (f = 0; f < FIELDS_COUNT; f++) {
		if (!field_ar[f].spec)
			continue;
		len += strlen(field_ar[f].spec);
		len ++; /* comma */
	}

	/* Add FORMAT_STR_EOL to the end of each line */
	len += (len / FORMAT_STR_LEN) * strlen(FORMAT_STR_EOL);

	/* '\n' + '\0' */
	p = malloc(len + 2);
	return p;
}

/*
 * Generate a string with all available format fields that can be
 * used with '-o' option. Used for '--help'
 *
 * Don't care about string overflow.
 */
void get_format_fields_def(char *s)
{
	unsigned f, found = 0;
	int n, len = 0;

	for (f = 0; f < FIELDS_COUNT; f++) {
		if (!field_ar[f].spec)
			continue;
		if (found++)
			strcpy(s++, ",");
		format_buf_add_eol(&s, &len);
		strcpy(s, field_ar[f].spec);
		n = strlen(field_ar[f].spec);
		s += n;
		len += n;
	}
}

/*
 * Generate a string with field set that is used in a given field_set array.
 * Don't care about string overflow.
 */
void get_format_fields(char *s, int *field_set)
{
	int id;
	unsigned f, found = 0;
	int n, len = 0;

	for (f = 0; f < FIELDS_COUNT; f++) {
		id = field_set[f];
		if (id == 0)
			break;
		if (!field_ar[id].spec)
			continue;
		if (found++)
			strcpy(s++, ",");
		format_buf_add_eol(&s, &len);
		strcpy(s, field_ar[id].spec);
		n = strlen(field_ar[id].spec);
		s += n;
		len += n;
	}
}

/*
 * Generate a string with format fields that can be
 * used with '-O' option. Used for '--help'.
 *
 * Don't care about string overflow.
 */
void get_format_fields_prel(char *s)
{
	int f, found, id;
	int num_fields_pre = pstat_mode?
		AR_SIZE(field_set_def_pre_pstat):
		AR_SIZE(field_set_def_pre);
	int num_fields_post = pstat_mode?
		AR_SIZE(field_set_def_post_pstat):
		AR_SIZE(field_set_def_post);

	found = 0;

	for (f = 0; f < num_fields_pre; f++) {
		id = pstat_mode?
			field_set_def_pre_pstat[f]:
			field_set_def_pre[f];
		if (id == 0)
			break;
		if (!field_ar[id].spec)
			continue;
		if (found++)
			strcpy(s++, ",");
		strcpy(s, field_ar[id].spec);
		s += strlen(field_ar[id].spec);
	}

	if (found++)
		strcpy(s++, ",");
	strcpy(s, "FMT");
	s += strlen("FMT");

	for (f = 0; f < num_fields_post; f++) {
		id = pstat_mode?
			field_set_def_post_pstat[f]:
			field_set_def_post[f];
		if (id == 0)
			break;
		if (!field_ar[id].spec)
			continue;
		strcpy(s++, ",");
		strcpy(s, field_ar[id].spec);
		s += strlen(field_ar[id].spec);
	}
}

void show_available_keys(void)
{
	char *p = alloc_format_buf();
	if (!p)
		exit_error("Can't allocate memory\n");
	get_format_fields_def(p);
	fprintf(stdout, "available keys: %s\n", p);
}

void usage()
{
	char *p;

	fprintf(stdout, "Usage: %sstat [options]\n", (pstat_mode) ? "p" : "vz");
	puts("Options:");
	puts("\t-h, -?\tThis help");
	puts("\t-d secs\tSpecifies delay");
	fprintf(stdout, "\t-p ID\tSelects %s to display info about (Up to 20 \n",
		(pstat_mode) ? "server" : "Container");
	fprintf(stdout, "\t       \t%s can be selected)\n",
		(pstat_mode) ? "servers" : "Containers");
	puts("\t-b\tBrief mode (mode 1, brief details)");
	puts("\t-v\tVerbose mode (mode 3, full details)");
	puts("\t-t\tText mode information for other programs parsing");
	puts("\t-l\tText mode information formatted as in interactive mode");
	puts("\t-m\tDisplay disc statistics for all fs types");
	puts("\t\t(by default only for ext2, ext3, reiserfs)");

	p = alloc_format_buf();
	if (!p)
		exit_error("Can't allocate memory\n");
	get_format_fields_def(p);
	fprintf(stdout, "\t-o FMT\tUser defined format:\n\t\t%s\n", p);

	get_format_fields_prel(p);
	fprintf(stdout, "\t-O FMT\tIdentical to:\n\t\t -o %s\n", p);

	get_format_fields(p, field_set_io);
	fprintf(stdout, "\t-i\tDisplay per-%s IO statistics. "
			"Identical to:\n\t\t -o %s\n",
			(pstat_mode) ? "server" : "Container", p);
	get_format_fields(p, field_set_net);
	fprintf(stdout, "\t-n\tDisplay per-%s network statistics. "
			"Identical to:\n\t\t -o %s\n",
			(pstat_mode) ? "server" : "Container", p);
	fprintf(stdout, "\t-s KEY\tSort per-%s statistics by KEY\n",
			(pstat_mode) ? "server" : "Container");
	fprintf(stdout, "\t-c B|K|M|G\tScale output to the specified units\n");


	free(p);
}

/*
 * return: 1 - show line
 *         0 - not show line
 */
int get_show_status(int flag)
{
	if (BIT_IS_SET(disp_flags, flag))
		return 1;
	return 0;
}

/*
 * Return an index of a given key in the field_ar array, if this
 * key is illegal, return -1
 */
int search_field_ar(const char *key)
{
	unsigned f;

	for (f = 0; f < AR_SIZE(field_ar); f++) {
		if (!field_ar[f].spec)
			continue;
		if (!strcmp(field_ar[f].spec, key))
			return f;
	}
	return -1;
}

/*
 * Here and below, msecs must have 'unsigned long' type,
 * because (signed long)tv_secs*1000 can be less than
 * zero depending on epoch value (1970, 2000, ...).
 */
static inline unsigned long tv2msecs(struct timeval *tv)
{
	return (tv->tv_usec / 1000 + tv->tv_sec * 1000);
}

static inline unsigned long tv_diff(struct timeval *old_tv,
					struct timeval *new_tv)
{
	return  (new_tv->tv_usec - old_tv->tv_usec) / 1000 +
		(new_tv->tv_sec - old_tv->tv_sec) * 1000;
}

#define __EINVAL	1
#define __ENOMEM	2

/*
 * Main parser calls this to save lists of keys for later.
 * Return ERROR if arg looks non-standard, and 0 otherwise.
 *
 * key1[=NAME1][:WIDTH1],key2[=NAME2]:[WIDTH2]
 *
 * key1  - column key
 * NAME  - column caption (used default name if not specified)
 * WIDTH - column width (used default width if not specified)
 */
int parse_o_option(char *s, char **pos, int preloaded)
{
	size_t dist;
	int field_item = -1;
	char buf[16];
	unsigned column = 0;
	int f, id;
	char *tmp;
	unsigned long width;
	int num_fields_pre = pstat_mode?
		AR_SIZE(field_set_def_pre_pstat):
		AR_SIZE(field_set_def_pre);
	int num_fields_post = pstat_mode?
		AR_SIZE(field_set_def_post_pstat):
		AR_SIZE(field_set_def_post);

	memset(field_set, 0, FIELDS_AR_SIZE);

	if (preloaded) {
		/* Insert prefix */
		for (f = 0; f < num_fields_pre; f++) {
			id = pstat_mode?
				field_set_def_pre_pstat[f]:
				field_set_def_pre[f];
			if (!id)
				break;
			if (field_ar[id].used++)
				continue;
			field_set[column++] = id;
		}
	}

	*pos = s;
	do {
		switch(*s) {
		case '\0':
			break;
		case ' ': case ',':
			field_item = -1;
			s++;
			*pos = s;
			break;
		case ':':
			if (field_item < 1)
				return __EINVAL;
			s++;
			*pos = s;
			dist = strcspn(s, ", =");
			if(dist >= sizeof(buf))
				return __EINVAL;

			width = strtoul(s, &s, 10);
			if (width == 0)
				return __EINVAL;
			field_ar[field_item].width = (int)width;
			break;
		case '=':
			if (field_item < 1)
				return __EINVAL;

			s++;
			*pos = s;
			if (*s == '\0') {
				/* 'key=' given */
				dist = 0;
				field_ar[field_item].head = "";
			} else {
				/* 'key=val' given */
				dist = strcspn(s, ", =:");
				if(dist >= sizeof(buf))
					return __EINVAL;

				/* Will never be freed!! */
				tmp = malloc(dist);
				if (!tmp)
					return __ENOMEM;

				strncpy(tmp, s, dist);	/* no '\0' on end */
			 	tmp[dist] = '\0';	/* fix that problem */
				s += dist;
				*pos = s;
				field_ar[field_item].head = tmp;
			}
			break;
		default:
			dist = strcspn(s, ", =:");
			if(dist >= sizeof(buf))
				return __EINVAL;
			strncpy(buf, s, dist);		/* no '\0' on end */
		 	buf[dist] = '\0';		/* fix that problem */
			field_item = search_field_ar(buf);
			if (field_item < 0)
				return __EINVAL;

			s += dist;
			*pos = s;
			if (field_ar[field_item].used++)
				break;
			field_set[column++] = field_item;
		}
	} while (*s != '\0' && column < FIELDS_COUNT);

	if (preloaded) {
		/* Insert postfix */
		for (f = 0; f < num_fields_post; f++) {
			id = pstat_mode?
				field_set_def_post_pstat[f]:
				field_set_def_post[f];
			if (!id)
				break;
			if (field_ar[id].used++)
				continue;
			if (column >= FIELDS_COUNT)
				break;
			field_set[column++] = id;
		}
	}

	return 0;
}

static int parse_scale_option(const char *key)
{
	if (!strcasecmp("bytes", key) ||
			!strcasecmp("b", key))
		scale = SCALE_BYTES;
	else if (!strcasecmp("kbytes", key) ||
			!strcasecmp("k", key))
		scale = SCALE_KBYTES;
	else if (!strcasecmp("mbytes", key) ||
			!strcasecmp("m", key))
		scale = SCALE_MBYTES;
	else if (!strcasecmp("gbytes", key) ||
			!strcasecmp("g", key))
		scale = SCALE_GBYTES;
	else
		return -1;

	return 0;
}

/*
 * Main parser calls this to set sort key in the text mode.
 * Return ERROR if arg looks non-standard, and 0 otherwise.
 */
int parse_s_option(char *s)
{
	unsigned f;

	/* Insert prefix */
	for (f = 0; f < AR_SIZE(field_ar); f++) {
		if (!field_ar[f].spec)
			continue;
		if (!strcmp(s, field_ar[f].spec)) {
			sort_mode = f;
			return 0;
		}
	}
	return -1;
}

void parse_options(int argc, char **argv)
{
	int veid, opt;
	char *pos;
	int ret;
	int fmt_used = 0;


	while ((opt = getopt(argc, argv, "d:p:o:O:s:c:inIAaltvbmh?")) > 0)
	{
		switch(opt)
		{
			case 'v':
				disp_flags = VL_VERBOSE;
				break;
			case 'b':
				disp_flags = VL_BRIEF;
				break;
			case 'd':
				if (sscanf(optarg, "%d", &sleep_time) != 1)
					exit_error("Invalid argument for -d"
						" %s\n", optarg);
				sleep_time *= 1000;
				timeout_specified = 1;
				break;
			case 't':
				has_color = 0;
				gen_nofmt_output = 1;
				gen_text_output = 1;
				break;
			case 'l':
				gen_text_output = 1;
				num_disp = 2;
				has_color = 0;
				break;
			case 'p':
				if (sscanf(optarg, "%d", &veid) != 1) {
					uuid_t u;
					if (uuid_parse(optarg, u))
						exit_error("Invalid argument for -p"
						" %s\n", optarg);
				}
				if (veids_specified == FILTER_MAX_CTID)
					exit_error("Too many Containers,"
						" max allowed is %d\n",
						FILTER_MAX_CTID);
				veids[veids_specified++] = strdup(optarg);
				break;
			case 'm':
				default_fs = 0;
				break;
			case 'i':
				if (net_mode)
					fprintf(stderr, "IO mode depricated. "
						"Using network mode\n");
				else {
					io_mode = 1;
					field_set = get_field_set(field_set_io);
				}
				break;
			case 'n':
				if (io_mode)
					fprintf(stderr, "Network mode "
						"depricated. Using IO mode\n");
				else {
					net_mode = 1;
					field_set = get_field_set(field_set_net);
				}
				break;
			case 'A':
			case 'a':
				io_stat_mode = 1;
				field_set = get_field_set(field_set_io_stat);
				break;
			case 'I':
				io_mode = 1;
				field_set = get_field_set(field_set_io);
				break;
			case 'O':
			case 'o':
				/* Don't allow to use two '-o' options */
				if (fmt_used)
					exit_error("Format option is already "
						"specified\n");
				if (optarg == NULL)
					exit_error("Format specification must"
						" follow -%c\n", (char)opt);
				ret = parse_o_option(optarg, &pos, opt == 'O');
				if (!ret) {
					/* OK */
					fmt_used = 1;
					break;
				}

				/* Error */

				fprintf(stderr, "Illegal field follow "
					" -%c\n", (char)opt);

				if (ret == __EINVAL) {
					char *p = optarg;

					/*
					 * highlight error position by '^'
					 */
					fprintf(stderr, "\t%s\n\t", optarg);
					while (p++ != pos)
						fprintf(stderr, " ");
					fprintf(stderr, "^\n");

					show_available_keys();
					exit_error(NULL);
				} else if (ret == __ENOMEM)
					exit_error("Not enough memory\n");

				/* Never reached */
				return;

			case 's':
				if (parse_s_option(optarg)) {
					fprintf(stderr, "Illegal sort key\n");

					show_available_keys();
					exit_error(NULL);
				}
				break;
			case 'c':
				if (parse_scale_option(optarg)) {
					fprintf(stderr, "Illegal scale option: %s\n",
							optarg);
					exit_error(NULL);
				}
				break;
			case '?':
			case 'h':
				usage();
				screen_close();
				exit(1);
			default:
				exit_error("Unknown option: %c\n", opt);
		}
	}
	if (optind < argc)
		exit_error("Unknown option: %s\n", argv[optind]);

	/*
	 * Backward compatibility - show veth statistics by default
	 * in text mode.
	 */
	if (gen_text_output)
		disp_flags |= MAKE_MASK(D_NET_VETH);
}

void parse_config()
{
	int i;
	FILE *f;
	char buf[128];

	f = fopen(CONFIG, "rt");
	if (!f) return;
	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		if (buf[0] == '#')
			continue;
		for (i = 0; config[i].opt != NULL; i++)
		{
			if (sscanf(buf, config[i].opt, config[i].param) == 1)
				break;
		}
	}
	fclose(f);
}

void parse_proc_vz_stats()
{
	FILE *f;
	char buf[512], name[32], *s, *ss, *ss2;
	int res, major, minor;
	unsigned int i, idx = 0;

	f = fopen(PROC_VZ_STATS, "rt");
	if (!f)
		return;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;

		if (sscanf(buf, "Version: %d.%d", &major, &minor) == 2)
			vz_stats_version = MKVERSION(major, minor);
		sscanf(buf, "cycles_per_jiffy: %lu", &cycles_per_jiffy);
		sscanf(buf, "jiffies_per_second: %lu", &jiffies_per_second);
		sscanf(buf, "VEs: %lu", &venum);
		sscanf(buf, "Processes avg: unint %lu %lu %lu",
			&unint_avg[0], &unint_avg[1], &unint_avg[2]);

		if (!strncmp(buf, "Processes: ", 11))
		{
			strncpy(proc_num_str, buf + 11, sizeof(proc_num_str));
			proc_num_str[sizeof(proc_num_str)-1] = 0;
		}

		if (!strncmp(buf, "Swap cache: ", 12))
		{
			strncpy(swap_cache_str, buf + 12, sizeof(swap_cache_str));
			swap_cache_str[sizeof(swap_cache_str)-1] = 0;
		}

		res = sscanf(buf, "ZONE %u %31s averages",
			&i,
			name);
		if (res == 2)
		{
			if (i >= sizeof(zone_info)/sizeof(zone_info_t)) continue;
			sscanf(buf, "ZONE%*d %*s averages: "
				"active %lu %lu %lu, "
				"inactive %lu %lu %lu, "
				"free %lu %lu %lu",
				&zone_info[i].active_avg[0],
				&zone_info[i].active_avg[1],
				&zone_info[i].active_avg[2],
				&zone_info[i].inactive_avg[0],
				&zone_info[i].inactive_avg[1],
				&zone_info[i].inactive_avg[2],
				&zone_info[i].free_avg[0],
				&zone_info[i].free_avg[1],
				&zone_info[i].free_avg[2]);
		}

		res = sscanf(buf, "%31s free", name);
		if (!res) continue;
		if (idx >= sizeof(zone_info)/sizeof(zone_info_t)) continue;

		s = strdup(buf);
		if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;

		/* find '(' in line and cut till the ')' */
		ss = strchr(s, '(');
		if (!ss) {free(s); continue;}
		*ss = 0;
		ss++;

		ss2 = strchr(ss, ')');
		if (!ss2) {free(s); continue;}
		*ss2 = 0;
		ss2++;

		if (zone_info[idx].name)
		{
			/* free previous strings */
			free(zone_info[idx].name);
			free(zone_info[idx].fragmentation);
		}

		zone_info[idx].fragmentation = strdup(ss);
		{
			size_t slen = strlen(s), ss2len = strlen(ss2);
			// only memmove could be used here since s and ss2 overlap
			memmove(s + slen, ss2, ss2len);
			s[slen + ss2len] = '\0';
		}
		zone_info[idx].name = s;

		idx++;
	}

	fclose(f);
}

void parse_proc_meminfo()
{
	FILE *f;
	char buf[128];

	memset(&meminfo, 0, sizeof(meminfo));
	f = fopen(PROC_MEMINFO, "rt");
	if (!f)
		// actually there will be divison by 0 later
		return;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		buf[sizeof(buf)-1] = 0;
		sscanf(buf, "MemTotal: %lu kB", &meminfo.mem_total);
		sscanf(buf, "MemFree: %lu kB", &meminfo.mem_free);
		sscanf(buf, "LowTotal: %lu kB", &meminfo.low_total);
		sscanf(buf, "LowFree: %lu kB", &meminfo.low_free);
		sscanf(buf, "HighTotal: %lu kB", &meminfo.high_total);
		sscanf(buf, "HighFree: %lu kB", &meminfo.high_free);
		sscanf(buf, "SwapTotal: %lu kB", &meminfo.swap_total);
		sscanf(buf, "SwapFree: %lu kB", &meminfo.swap_free);

		sscanf(buf, "Buffers: %lu kB", &meminfo.buffers);
		sscanf(buf, "Cached: %lu kB", &meminfo.cached);
		sscanf(buf, "Active: %lu kB", &meminfo.active);
		sscanf(buf, "Inactive: %lu kB", &meminfo.inactive);
	}
	if (meminfo.low_total == 0)
		meminfo.low_total = meminfo.mem_total;
	if (meminfo.low_free == 0)
		meminfo.low_free = meminfo.mem_free;

	fclose(f);
}

void parse_proc_vz_mmperf()
{
	FILE *f;
	char buf[1024];
	int major, minor;

	f = fopen(PROC_VZ_MMPERF, "rt");
	if (!f) return;
	memcpy(&swapout_prev, &swapout, sizeof(swapout));
	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		if (sscanf(buf, "Version: %d.%d", &major, &minor) == 2)
			vz_mmperf_version = MKVERSION(major, minor);
		PARSE_KSTAT_FUNC(buf, "swap_out:", swapout);
	}
	fclose(f);
}

void parse_proc_uptime()
{
	FILE *f;
	unsigned long tmp;

	f = fopen(PROC_UPTIME, "rt");
	if (!f)
		// actually there will be divison by 0 later
		return;

	if (fscanf(f, "%lu.%lu %lu.%lu", &uptime_secs, &tmp, &idle_secs, &tmp) != 4)
	{
		uptime_secs = 0;
		idle_secs = 0;
	}

	fclose(f);
}

void parse_proc_loadavg()
{
	FILE *f;
	unsigned long res, avg1, avg11, avg2, avg22, avg3, avg33;

	f = fopen(PROC_LOADAVG, "rt");
	if (!f)
		// actually there will be divison by 0 later
		return;

	memset(loadavg, 0, sizeof(loadavg));
	res = fscanf(f, "%lu.%lu %lu.%lu %lu.%lu",
		&avg1, &avg11, &avg2, &avg22, &avg3, &avg33);
	if (res == 6)
	{
		loadavg[0] = avg1 + avg11/100.0;
		loadavg[1] = avg2 + avg22/100.0;
		loadavg[2] = avg3 + avg33/100.0;
	}

	fclose(f);
}

static inline diskstat_t *get_diskstat(unsigned int major, unsigned int minor)
{
	diskstat_t *ret;
	int i;

	for (i = 0; i < MAX_DISK; i++) {
		ret = &diskstat[i];
		/* New element */
		if (!ret->major && !ret->minor)
			break;
		if (ret->major == major && ret->minor == minor)
			break;
	}
	if (i == MAX_DISK)
		ret = NULL;
	return ret;
}

void parse_disk_stat(char *line)
{
	char *p, *ep;
	unsigned long long io[3];
	diskstat_t *d;
	unsigned int major, minor;

	p = line;
	ep = line + strlen(line);
	while (p < ep) {
		if (sscanf(p, "(%u,%u):(%llu,%*u,%llu,%*u,%llu) ",
		    &major, &minor, &io[0], &io[1], &io[2]) != 5)
			break;

		d = get_diskstat(major, minor);
		if (!d)
			continue;

		d->major = major;
		d->minor = minor;
		d->delta_rblk = io[1] - d->rblk;
		d->delta_wblk = io[2] - d->wblk;
		d->rblk = io[1];
		d->wblk = io[2];
		if ((p = strchr(p, ' ')) == NULL)
			break;
		p++;
	}
}

static void calc_disk_rate(unsigned long msecs)
{
	int i;

	for (i = 0; i < MAX_DISK; i++)
	{
		if (!diskstat[i].major && !diskstat[i].minor)
			break;
		if (msecs) {
			diskstat[i].delta_rblk /= ((float)msecs/1000.0);
			diskstat[i].delta_wblk /= ((float)msecs/1000.0);
		} else {
			diskstat[i].delta_rblk = 0.0;
			diskstat[i].delta_wblk = 0.0;
		}
	}
}

static inline int is_visible_disk(char * name)
{
	char device[64], * slash;
	struct stat sb;

	while ((slash = strchr(name, '/')))
		*slash = '!';

	snprintf(device, sizeof(device), "/sys/block/%s/device", name);

	return !lstat(device, &sb);
}

void parse_proc_diskstats()
{
	FILE *f;
	char buf[128];
	char name[33];
	unsigned int major, minor;
	unsigned int in_flight;
	unsigned long long io[3];
	diskstat_t *d;
	static struct timeval old_tv = {0, 0};
	struct timeval tv;
	unsigned long msecs;
	int res;

	f = fopen(PROC_DISKSTATS, "rt");
	if (!f)
		return;
	else
		has_proc_diskstats = 1;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		res = sscanf(buf,"%u %u %32s %*u %*u %llu %*u %*u "
				"%*u %llu %*u %u %*u %*u\n",
				&major, &minor, name, &io[1], &io[2], &in_flight);
		if (res != 6)
				continue;

		if (!is_visible_disk(name))
			continue;

		d = get_diskstat(major, minor);
		if (!d)
			continue;

		d->major = major;
		d->minor = minor;
		d->delta_rblk = io[1] - d->rblk;
		d->delta_wblk = io[2] - d->wblk;
		d->rblk = io[1];
		d->wblk = io[2];
	}
	fclose(f);

	gettimeofday(&tv, NULL);
	msecs = tv_diff(&old_tv, &tv);

	calc_disk_rate(msecs);

	old_tv = tv;
}

int parse_proc_stat(char *fname, int allow_parse_disk_stat,
			 	unsigned long *user,
				unsigned long *nice,
				unsigned long *system,
				unsigned long *idle,
				unsigned long *iowait)
{
	FILE *f;
	char buf[128];

	f = fopen(fname, "rt");
	if (!f)
		return -1;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f))
			break;
		if (!strncmp(buf, "cpu ", 4))
			sscanf(buf, "cpu %lu %lu %lu %lu %lu", user, nice, system, idle, iowait);
		else if (allow_parse_disk_stat && !strncmp(buf, "disk_io: ", 9))
				parse_disk_stat(buf + 9);
	}
	fclose(f);

	return 0;
}

void parse_ve_proc_stat(s_line_t *vestat)
{
	char fname[PATH_MAX];
	unsigned long user = 0, nice = 0, system = 0, idle = 0, iowait = 0;

	snprintf(fname, sizeof(fname), PROC_VZ_STAT,
			"machine.slice/", vestat->veid ? vestat->ve_uuid : "");
	if (access(fname, F_OK))
		snprintf(fname, sizeof(fname), PROC_VZ_STAT,
				"", vestat->veid ? vestat->ve_uuid : "");

	if (parse_proc_stat(fname, 0, &user, &nice, &system, &idle, &iowait) < 0)
		return;

	vestat->cpu[CPU_USER] = user;
	vestat->cpu[CPU_NICE] = nice;
	vestat->cpu[CPU_SYSTEM] = system;
	vestat->cpu[CPU_IDLE] = idle;
	vestat->cpu[CPU_IOWAIT] = iowait;
}

void parse_ve_proc_delay(s_line_t *vestat)
{
	char fname[PATH_MAX];
	char buf[128];
	unsigned long swapin_count = 0;
	FILE *f;

	snprintf(fname, sizeof(fname), PROC_VZ_DELAY,
			"machine.slice/", vestat->veid ? vestat->ve_uuid : "");
	if (access(fname, F_OK))
		snprintf(fname, sizeof(fname), PROC_VZ_DELAY,
				"", vestat->veid ? vestat->ve_uuid : "");

	f = fopen(fname, "rt");
	if (!f)
		return;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f))
			break;
		if (sscanf(buf, "swapin_count %lu", &swapin_count) == 1)
			break;
	}
	fclose(f);

	vestat->io[IO_SWAPIN_COUNT] = swapin_count;
}

static int parse_proc_vmstat(unsigned long *swapin, unsigned long *swapout)
{
	FILE *fp = fopen(PROC_VMSTAT, "r");
	if (!fp)
		return -1;
	while (!feof(fp))
	{
		char buf[128] = {};
		if (!fgets(buf, sizeof(buf), fp))
			break;
		sscanf(buf, "pswpin %lu\n", swapin);
		sscanf(buf, "pswpout %lu\n", swapout);
	}
	fclose(fp);
	return 0;
}

void parse_node_proc_stat()
{
	unsigned long user = 0, nice = 0, system = 0, idle = 0, iowait = 0;
	unsigned long si = 0, so = 0;
	unsigned long msecs;
	struct timeval tv;

	static unsigned long user_prev, nice_prev, system_prev,
				idle_prev, iowait_prev;
	static struct timeval old_tv;
	static unsigned long old_si, old_so;


	if (parse_proc_stat(PROC_STAT, !has_proc_diskstats, &user, &nice, &system, &idle, &iowait) < 0)
		return;
	if (parse_proc_vmstat(&si, &so))
		return;

#define CALC_DIFF(x, y) (((x)-(y)) > 2000000000 ? 0 : \
				(((x)-(y)) > 0 ? ((x)-(y)) : 0))
	total_cpu_clk = user + nice + system + idle + iowait;

	user_j = CALC_DIFF(user, user_prev);
	nice_j = CALC_DIFF(nice, nice_prev);
	system_j = CALC_DIFF(system, system_prev);
	idle_j = CALC_DIFF(idle, idle_prev);
	iowait_j = CALC_DIFF(iowait, iowait_prev);

	user_prev = user;
	nice_prev = nice;
	system_prev = system;
	idle_prev = idle;
	iowait_prev = iowait;
#undef CALC_DIFF

	gettimeofday(&tv, NULL);
	msecs = tv_diff(&old_tv, &tv);

	if (msecs) {
		swap_in = (float)(si - old_si) * page_size /
				(msecs/1000.0) / 1024 / 1024;
		swap_out = (float)(so-old_so) * page_size /
				(msecs/1000.0) / 1024 / 1024;
	} else {
		/* paranoic */
		swap_in = 0.0;
		swap_out = 0.0;
	}

	if (!has_proc_diskstats)
		calc_disk_rate(msecs);

	old_tv = tv;
	old_si = si;
	old_so = so;
}

void parse_proc_vz_latency()
{
	FILE *f;
	char buf[1024];
	cycles_t time;
	int averages = 0, res;
	unsigned long num;

	f = fopen(PROC_VZ_LATENCY, "rt");
	if (!f)
		return;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		if (!strncmp(buf, "Averages", 8)) averages = 1;

#define PARSE_LAT(str, name)					\
		if (!averages)					\
		{						\
			res = sscanf(buf, str " %Lu %Lu %lu\n",	\
				&name.max,			\
				&time,				\
				&num);				\
			if (res && num != name.num)		\
			{					\
				/* save old values */		\
				name.prev_num = name.num;	\
				name.prev_time = name.time;	\
				name.num = num;			\
				name.time = time;		\
			}					\
		}						\
		else						\
			sscanf(buf, str " %Lu %Lu %Lu\n",	\
				&name.avg[0],			\
				&name.avg[1],			\
				&name.avg[2]);

		PARSE_LAT("scheduling:", sched)
		PARSE_LAT("alocatomic:", a0)
		PARSE_LAT("aloclow:", k0)
		PARSE_LAT("alochigh:", u0)
		PARSE_LAT("aloclowmp:", k1)
		PARSE_LAT("alochighmp:", u1)
		PARSE_LAT("swap_in:", swapin)

#undef PARSE_LAT
	}
	fclose(f);
	vz_latency_version = 1;
}

void parse_proc_slabinfo()
{
	FILE *f;
	int major, minor;
	unsigned long active_objs, num_objs, objsize;
	unsigned long active_slabs, num_slabs, gfporder;
	char buf[128], name[32];
	char *cp;
	int res;

	f = fopen(PROC_SLABINFO, "rt");
	if (!f)
		return;

	if (!fgets(buf, sizeof(buf), f))
		goto out;
	res = sscanf(buf, "slabinfo - version: %d.%d%*s\n", &major, &minor);
	if (res != 2)
		goto out;

	proc_slabinfo_version = MKVERSION(major, minor);

	slab_total = slab_active = 0;
	slab_inode = slab_dentry = slab_bh = slab_pb = 0;
	while (!feof(f)) {
		if (!fgets(buf, sizeof(buf), f))
			break;

		switch (proc_slabinfo_version) {
		case MKVERSION(1, 1):
			res = sscanf(buf, "%31s %lu %lu %lu %lu %lu %lu",
				name, &active_objs, &num_objs, &objsize,
				&active_slabs, &num_slabs, &gfporder);
			if (res != 7) continue;
			break;
		case MKVERSION(2, 1):
			/*
			 * cachep->objsize (2.6.9) ===
			 *	cachep->buffer_size (2.6.18)
			 * fall through ...
			 */
		case MKVERSION(2, 0):
			res = sscanf(buf, "%17s %lu %lu %lu %*u %ld"
				" : tunables %*u %*u %*u"
				" : slabdata %lu %lu %*u",
				name, &active_objs, &num_objs, &objsize,
				&gfporder, &active_slabs, &num_slabs);
			if (res != 7) continue;
			break;
		default:
			goto out;
		}

		gfporder *= page_size;
		slab_total += num_slabs*gfporder;
		slab_active += active_slabs*gfporder;
		if (!strcmp(name, "dentry_cache"))
			slab_dentry = num_slabs*gfporder;
		else if (!strcmp(name, "buffer_head"))
			slab_bh = num_slabs*gfporder;
		else if (!strcmp(name, "page_beancounter"))
			slab_pb = num_slabs*gfporder;
		else {
			int len;

			len = strlen(name);
			if (len < 11)
				continue;
			cp = name + len - 11;
			if (!strcmp(cp, "inode_cache"))
				slab_inode += num_slabs*gfporder;
		}
	}
out:
	fclose(f);
}

void parse_netdev()
{
	FILE *f;
	char buf[4096], name[32], *s;
	unsigned long rx, pr, tx, pt, tmp, res, idx;
	unsigned long msecs;
	ifstat_t *ifs;
	struct timeval tv;
	static struct timeval old_tv;

	f = fopen(PROC_NET_DEV, "rt");
	if (!f)
		return;

	idx = 0;
	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;

		s = strchr(buf, ':');
		if (!s) continue;
		*s = 0; s++;

		res = sscanf(buf, "%31s", name);
		if (res != 1) continue;

		res = sscanf(s,
			"%lu %lu %lu %lu %lu %lu %lu %lu "
			"%lu %lu %lu %lu %lu %lu %lu %lu ",
			&rx, &pr, &tmp, &tmp, &tmp, &tmp, &tmp, &tmp,
			&tx, &pt, &tmp, &tmp, &tmp, &tmp, &tmp, &tmp);
		if (res != 16) continue;
		if (idx >= sizeof(ifstats)/sizeof(ifstat_t)) break;
		if (!strncmp(buf, "venet", 5)) continue;

		ifs = &ifstats[idx];
		ifs->rx = rx-ifs->old_rx;
		ifs->tx = tx-ifs->old_tx;
		ifs->pr = pr-ifs->old_pr;
		ifs->pt = pt-ifs->old_pt;
		ifs->old_rx = rx;
		ifs->old_tx = tx;
		ifs->old_pr = pr;
		ifs->old_pt = pt;
		if (ifs->name)
			free(ifs->name);
		ifs->name = strdup(name);
		idx++;
	}
	fclose(f);
	ifcount = idx;

	gettimeofday(&tv, NULL);
	msecs = tv_diff(&old_tv, &tv);
	old_tv = tv;

	for (idx=0; idx<ifcount; idx++)
	{
		ifs = &ifstats[idx];
		if (msecs) {
			ifs->rx = ifs->rx/(msecs/1000.0);
			ifs->tx = ifs->tx/(msecs/1000.0);
			ifs->pr = ifs->pr/(msecs/1000.0);
			ifs->pt = ifs->pt/(msecs/1000.0);
		} else {
			/* paranoic */
			ifs->rx = 0;
			ifs->tx = 0;
			ifs->pr = 0;
			ifs->pt = 0;
		}
	}
}

int check_level_le(float val, float level1, float level2)
{
	if (level2 != -1 && val <= level2)
		return C_RED;
	else if (level1 != -1 && val <= level1)
		return C_YELLOW;
	return C_NORMAL;
}

int check_level_ge(float val, float level1, float level2)
{
	if (level2 != -1 && val >= level2)
		return C_RED;
	else if (level1 != -1 && val >= level1)
		return C_YELLOW;
	return C_NORMAL;
}

static int disk_get_id_by_mnt(char *mnt)
{
	int i;

	for(i = 0; i < MAX_DISK; i++)
	{
		if (diskinfo[i].mnt == NULL)
			return i;
		if (!strcmp(diskinfo[i].mnt, mnt))
			return i;
	}
	return i;
}

static void disk_clear_inuse()
{
	int i;
	for(i = 0; i < MAX_DISK; i++)
		diskinfo[i].inuse = 0;
}

char *parse_var(char *var)
{
	char *sp, *ep;

	if (var == NULL)
		return NULL;
	sp = var;
	while (*sp && (*sp == '"' || isspace(*sp))) sp++;
	ep = var + strlen(var) - 1;
	while (ep > sp && (*ep == '"' || isspace(*ep))) *ep-- = 0;
	return strdup(sp);
}

static void parse_ve_conf(s_line_t *vestat)
{
	FILE *f;
	char buf[1024];
	char data[1024];

	sprintf(buf, VE_CONF_DIR, vestat->ve_uuid);
	f = fopen(buf, "r");
	if (!f)
		return;
	while (!feof(f)) {
		if (fgets(buf, sizeof(buf), f) == NULL)
			break;
		if (sscanf(buf, "HOSTNAME=%1024s", data) == 1) {
			if (vestat->hostname) free(vestat->hostname);
			vestat->hostname = parse_var(data);
		}
	}
	fclose(f);
	return;
}

int check_fs(char **fs_names, char *fs)
{
	int i;

	for (i = 0; fs_names[i] != NULL; i++)
		if (strncmp(fs, fs_names[i], strlen(fs_names[i])) == 0)
			return 1;
	return 0;
}

void parse_mounts()
{
	FILE *f;
	char buf[128], name[128] = {}, dev[128] = {}, fs[128] = {};
	struct statfs st;
	int idx = 0;
	static char *skip_fs_names[] = {
		"vzfs", "proc", "tmpfs", "shmfs", "devfs", "iso9660",
		"devpts", "autofs", "usbdevfs", "binfmt", NULL};

	f = fopen(PROC_MOUNTS, "rt");
	if (!f)
		return;
	disk_clear_inuse();
	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f))
			break;
		if (sscanf(buf, "%127s %127s %127s", dev, name, fs) != 3)
			continue;
		if (!strcmp(dev, "none"))
			continue;
		if (check_fs(skip_fs_names, fs))
			continue;
		if (default_fs && !check_fs(default_fs_names, fs))
			continue;
		if (statfs(name, &st))
			continue;
		if (!st.f_blocks)
			continue;
		idx = disk_get_id_by_mnt(name);
		if (idx == MAX_DISK ) /* array diskinfo is full */
			continue;
		if (diskinfo[idx].mnt == NULL) /* New element */
		{
			char *p;
			diskinfo[idx].mnt = strdup(name);
			if ((p = strrchr(dev, '/')) == NULL)
				p = dev;
			else
				p++;
			diskinfo[idx].dev = strdup(p);
		}
		diskinfo[idx].inuse = 1;
		diskinfo[idx].disk_total = (unsigned long long) st.f_blocks *
			st.f_bsize;
		diskinfo[idx].disk_free = (unsigned long long) st.f_bfree *
			st.f_bsize;
		diskinfo[idx].inodes_total = st.f_files;
		diskinfo[idx].inodes_free = st.f_ffree;
	}
	fclose(f);
}

/*
 * Parse VE's IO statistics file
 */
static void parse_proc_ioacct(s_line_t *vestat)
{
	FILE *f;
	char buf[128];
	char fname[256];
	int step;
	unsigned long long dirty;

	/*
	 * Generate file name, return if buf is too small ...
	 */
	int n = snprintf(fname, sizeof(fname), PROC_VZ_IOACCT, vestat->ve_uuid);
	if (n < 0 || n >= (int)sizeof(fname))
		return;
	fname[sizeof(fname) - 1] = '\0';

	f = fopen(fname, "rt");
	if (!f)
		return;

	step = 0;
	vestat->io[IO_SYNC] = 0;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f))
			break;

		/* Step #1, get sync stats */
		/* Split to steps just an optimization */
		if (step == 1) {
			unsigned long long sync = 0;
			if (sscanf(buf, " %*ssyncs_total %Lu", &sync))
				vestat->io[IO_SYNC] += sync;
			continue;
		}

		/* Step #0, get general stats */
		sscanf(buf, " read %Lu", &vestat->io[IO_READ]);
		sscanf(buf, " write %Lu", &vestat->io[IO_WRITE]);
		sscanf(buf, " dirty %Lu", &dirty);
		sscanf(buf, " cancel %Lu", &vestat->io[IO_CANCEL]);
		if (sscanf(buf, " missed %Lu", &vestat->io[IO_MISSED]))
			step++;
	}
	vestat->io[IO_DIRTY] = dirty - vestat->io[IO_WRITE];
	fclose(f);
}

static void parse_proc_bc_meminfo(s_line_t *vestat)
{
	FILE *f;
	char buf[128];
	char fname[256];

	int n = snprintf(fname, sizeof(fname), PROC_VZ_MEMINFO, vestat->ve_uuid);
	if (n < 0 || n >= (int)sizeof(fname))
		return;
	fname[sizeof(fname) - 1] = '\0';

	f = fopen(fname, "rt");
	if (!f)
		return;

	while (!feof(f))
	{
		if (!fgets(buf, sizeof(buf), f)) break;
		buf[sizeof(buf)-1] = 0;
		sscanf(buf, "MemTotal: %lu kB", &vestat->mem_total);
		sscanf(buf, "MemFree: %lu kB", &vestat->mem_free);
		sscanf(buf, "SwapTotal: %lu kB", &vestat->swap_total);
		sscanf(buf, "SwapFree: %lu kB", &vestat->swap_free);
	}

	fclose(f);
}

static void parse_proc_iostat(s_line_t *vestat)
{
	FILE *f;
	char buf[128];
	char fname[256];
	unsigned long long dummy, wait_tm, used_tm;
	unsigned long long req_completed, transferred;

	int n = snprintf(fname, sizeof(fname), PROC_VZ_IOSTAT, vestat->ve_uuid);
	if (n < 0 || n >= (int)sizeof(fname))
		return;
	fname[sizeof(fname) - 1] = '\0';

	f = fopen(fname, "rt");
	if (!f)
		return;

	vestat->io[IO_WAIT_TM] = 0;
	vestat->io[IO_USED_TM] = 0;
	vestat->io[IO_REQ_COMPLETED] = 0;
	vestat->io[IO_TRANSFERRED_SEC] = 0;

	while (!feof(f))
	{
		char c;
		if (!fgets(buf, sizeof(buf), f))
			break;
/*
1	disk			Disk device name, e.g. sda or hda
2	ub_id			Beancounter id
3	state			Possible values: I - idle, W - wait, A - active (see below)
4	busy_queues		The number of queues with requests (see below)
5	on_ dispatch		The number of requests dispatched to a hardware
6	activations_count	The number of switches to active state
7	wait_time		Total time in waiting state in milliseconds
8	used_time		Total time in active state in milliseconds.
9	requests_completed	The number of completed requests
10	sectors_ transferred	The number of 512 sectors transferred (includes both re
*/

		wait_tm = used_tm = req_completed = transferred = 0;
		sscanf(buf, "%*s %*s %c %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
				&c, &dummy, &dummy, &dummy,
				&wait_tm,
				&used_tm,
				&req_completed,
				&transferred);
		vestat->io[IO_WAIT_TM] += wait_tm;
		vestat->io[IO_USED_TM] += used_tm;
		vestat->io[IO_REQ_COMPLETED] += req_completed;
		vestat->io[IO_TRANSFERRED_SEC] += transferred;
	}

	fclose(f);
}

static int vzctlfd = -1;

int get_vzctlfd()
{
	if (vzctlfd != -1)
		return vzctlfd;

	vzctlfd = open(VZCTLDEV, O_RDWR);

	return vzctlfd;
}

void close_vzctlfd()
{
	if (vzctlfd != -1) {
		close(vzctlfd);
		vzctlfd = -1;
	}
}

static void get_io(const char *id, const char *name, unsigned long long *limit)
{
	FILE *fp;
	char f[512];

	snprintf(f, sizeof(f), "/sys/fs/cgroup/beancounter/%s/beancounter.%s.speed",
			id, name);
	fp = fopen(f, "r");
	if (fp == NULL)
		return;
	if (fscanf(fp, "%llu", limit) != 1)
		*limit = 0;
	fclose(fp);
}

static void get_iolimits(s_line_t *vestat)
{
	get_io(vestat->ve_uuid, "iolimit", &vestat->io_limit);
	get_io(vestat->ve_uuid, "iopslimit", &vestat->iops_limit);
}

void parse_proc_io()
{
	unsigned long msecs;
	struct timeval tv;
	int idx = 0;

	while (idx >= 0)
	{
		s_line_t *vestat;

		vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;

		parse_proc_ioacct(vestat);
		parse_proc_iostat(vestat);
		parse_ve_proc_delay(vestat);
		get_iolimits(vestat);

		gettimeofday(&tv, NULL);
		msecs = tv2msecs(&tv);
		vestat->io_msecs = msecs;
	}
}

#define FREE_STAT(x)		\
	if (x != NULL) {	\
		free(x);	\
		x = NULL;	\
	}


void parse_net_stats(void)
{
	int i, cl;
	struct vzctl_tc_get_stat_list velist = {NULL, 0};
	struct vzctl_tc_get_stat netstat;
	int idx, ret;
	s_line_t *vestat;
	unsigned long long in, out, in_p, out_p;
	unsigned long msecs;
	struct timeval tv;
	int fd;

	fd = get_vzctlfd();
	if (fd == -1)
		return;
	if ((ret = get_ve_list(&velist)) < 0)
		goto out;

	gettimeofday(&tv, NULL);

	msecs = tv2msecs(&tv);

	for (i = 0; i < velist.length; i++)
	{
		if ((ret = get_ve_stats(velist.list[i], &netstat)))
			goto out;

		/* Find appropriate VEstat */
		idx = 0;
		while (1)
		{
			vestat = vzstatcpu_get_ve_stats(idx++, 1);
			if (!vestat)
				break;

			if (vestat->veid != netstat.veid)
				continue;

			/* VE is found, now summarize all classes stats */
			in = out = in_p = out_p = 0;

			for (cl = 0; cl < netstat.length; cl++)
			{
				in    += netstat.incoming[cl];
				out   += netstat.outgoing[cl];
				in_p  += netstat.incoming_pkt[cl];
				out_p += netstat.outgoing_pkt[cl];
			}
			vestat->net.total[NET_RX]      = in;
			vestat->net.total[NET_TX]      = out;
			vestat->net.total[NET_RX_PKTS] = in_p;
			vestat->net.total[NET_TX_PKTS] = out_p;

			vestat->net.msecs = msecs;

			break;
		}

		FREE_STAT(netstat.incoming);
		FREE_STAT(netstat.outgoing);
		FREE_STAT(netstat.incoming_pkt);
		FREE_STAT(netstat.outgoing_pkt);
	}
out:
	if (velist.list)
		free(velist.list);
	return;
}

void parse_mem_stats()
{
	s_line_t *vestat;
	int idx = 0;

	while (idx >= 0)
	{
		vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;
		parse_proc_bc_meminfo(vestat);
	}
}

void parse_vz_stats()
{
	s_line_t *vestat;
	int idx = 0;

	while (idx >= 0)
	{
		vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;
		parse_ve_proc_stat(vestat);
	}
}

void parse_slm_stats()
{
	s_line_t *vestat;
	unsigned long long slm_quality_total = 0;
	int idx = 0;

	while (idx >= 0)
	{
		vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;

		if (parse_slm_info(vestat->veid, &vestat->slm_info))
			continue;
		vestat->slm = 1;
		vestat->privvmpages.held = vestat->slm_info.mem.inst_usage;
		vestat->kmemsize.held = vestat->slm_info.lowmem.inst_usage * page_size;

		slm_quality_total += vestat->slm_info.mem.quality;
	}

	idx = 0;
	while (idx >= 0) {
		vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;

		vestat->slm_info.oomlimit = (float) (vestat->slm_info.mem.quality) * meminfo.mem_total /
					slm_quality_total;
	}
}

static void parse_proc_task_stat(s_line_t *vestat)
{
	FILE *fp;
	char buf[512];
	char *s;
	int num;
	unsigned long long utime, stime;

	if (vestat->prl_vm_pid == 0)
		return;
	sprintf(buf, "/proc/%d/stat", vestat->prl_vm_pid);
	fp = fopen(buf, "rt");
	if (fp == NULL)
		return;
	if (fgets(buf, sizeof(buf), fp) == NULL) {
		fclose(fp);
		return;
	}
	s = strchr(buf, ')');
	if (s != NULL) {
		s += 1;
		num = sscanf(s,
			" %*c " /* state */
			"%*d %*d %*d %*d %*d " /* ppid pgrp session tty tpgid */
			"%*u %*u %*u %*u %*u " /* flags min_flt cmin_flt maj_flt maj_flt */
			"%Lu %Lu ",  /* utime stime */
			&utime, &stime);

		if (num == 2) {
			vestat->cycles = total_cpu_clk;
			vestat->used_time = utime + stime;
		}
	}

	fclose(fp);
}

void vzstatcpu_proc_stat()
{
	int idx = 0;

	while (idx >= 0)
	{
		s_line_t *vestat = vzstatcpu_get_ve_stats(idx++, 1);
		if (!vestat)
			break;
		if (vestat->env_type != VM_TYPE)
			continue;
		parse_proc_task_stat(vestat);
	}
}

void update_vestats()
{
	FILE *f;
	int idx = 0;

	vzstatcpu_begin_stats();

	if (cg_find_ves())
		return;

	f = fopen(PROC_VZ_VESTAT, "rt");
	if (f)
	{
		vzstatcpu_parse_vestat(f);
		fclose(f);
	}

	f = fopen(PROC_UB, "rt");
	if (f)
	{
		vzstatcpu_parse_ub(f);
		fclose(f);
	}

	f = fopen(PROC_VZ_VEINFO, "rt");
	if (f)
	{
		vzstatcpu_parse_veinfo(f);
		fclose(f);
	}


	vzstatcpu_parse_cpu();	
	parse_proc_io();
	parse_net_stats();
	parse_slm_stats();
	parse_mem_stats();
	parse_vz_stats();
	parse_uuid();
	if (pstat_mode)
		vzstatcpu_proc_stat();

	vzstatcpu_update_stats();

	vmnum = 0;
	while (idx >= 0)
	{
		s_line_t *vestat;

		vestat = vzstatcpu_get_ve_stats(idx++, 2);
		if (!vestat) break;

		vestat->maxlat_st = check_level_ge((float)vestat->maxlat/
			CYCLES_PER_SEC*1000.0,
			cpu_latmax1_fail_level,
			cpu_latmax2_fail_level);
		vestat->fcnt_st = check_level_ge(vestat->fcnt_delta, -1,
			ve_delta_fcnt_fail_level);

		vestat->status = vestat->maxlat_st + vestat->fcnt_st;

		if (vestat->env_type == VM_TYPE)
			vmnum++;
	}
}

void parse_base_proc_files()
{
	parse_proc_uptime();
	parse_proc_loadavg();
	parse_proc_meminfo();
	/* try /proc/diskstats first */
	parse_proc_diskstats();
	parse_node_proc_stat();
	parse_proc_slabinfo();
	parse_netdev();
	parse_mounts();
	parse_proc_vz_stats();
	parse_proc_vz_latency();
	parse_proc_vz_mmperf();
	update_vestats();
}

void show_current_time()
{
	time_t realseconds;
	struct tm *realtime;

	time(&realseconds);
	realtime = localtime(&realseconds);
	printe(" %2d:%02d%s",
		realtime->tm_hour%12 ? realtime->tm_hour%12 : 12,
		realtime->tm_min, realtime->tm_hour > 11 ? "pm" : "am");
}

void show_uptime()
{
	unsigned long updays, uphours, upminutes;

	updays = (int) uptime_secs / (60*60*24);
	printe("up ");
	if (updays)
		printe("%d day%s, ", updays, (updays != 1) ? "s" : "");
	upminutes = (int) uptime_secs / 60;
	uphours = upminutes / 60;
	uphours = uphours % 24;
	upminutes = upminutes % 60;
	if(uphours)
		printe("%2d:%02d", uphours, upminutes);
	else
		printe("%d min", upminutes);
}

void show_users()
{
	unsigned long numuser = 0;
	struct utmp *utmpstruct;

	setutent();
	while ((utmpstruct = getutent())) {
	if ((utmpstruct->ut_type == USER_PROCESS) &&
			(utmpstruct->ut_name[0] != '\0'))
		numuser++;
	}
	endutent();

	printe("%2d user%s", numuser, numuser == 1 ? "" : "s");
}

void show_loadavg()
{
	if (gen_nofmt_output)
	{
		printf("loadavg\t\t%d %d %d",
			(int)loadavg[0],
			(int)loadavg[1],
			(int)loadavg[2]);
	}
	else
	{
		printe("load average: ");
		printc(check_level_ge(loadavg[0], lavg_fail_level, -1),
		       	"%.2f", loadavg[0]);
		printe(", ");
		printc(check_level_ge(loadavg[1], lavg_fail_level, -1),
		       	"%.2f", loadavg[1]);
		printe(", ");
		printc(check_level_ge(loadavg[2], lavg_fail_level, -1),
		       	"%.2f", loadavg[2]);
	}
}

void show_overall_stat()
{
	if (!get_show_status(D_LOADAVG))
		return;

	if (!gen_nofmt_output)
	{
		show_current_time();
		printe(", ");
		show_uptime();
		printe(", ");
		show_users();
		printe(", ");
	}
	show_loadavg();
	printe("\n");
}

void show_proc_stat()
{
	int procs, running, sleeping, stopped, unint, zombie, dead;
	int run_st, unint_st;

	if (!get_show_status(D_PROC_INFO))
		return;
	procs = running = sleeping = stopped = unint = zombie = 0;
	if (vz_stats_version < MKVERSION(2, 6)) {
		if (sscanf(proc_num_str, "run %d, sleep %d, "
			"unint %d, zombie %d, stop %d",
			&running, &sleeping,
			&unint,	&zombie, &stopped) != 5)
			return;
		dead = 0;
	} else {
		if (sscanf(proc_num_str, "R %d, S %d, "
			"D %d, Z %d, T %d, X %d",
			&running, &sleeping,
			&unint,	&zombie, &stopped, &dead) != 6)
			return;
	}
	procs = running + sleeping + unint + zombie + stopped + dead;
	if (avg_mode)
		unint = unint_avg[0];
	if (gen_nofmt_output)
	{
		printf("CTNum\t\t%lu\n", venum - 1);
		if (pstat_mode)
			printf("VMNum\t\t%lu\n", vmnum);
		printf("procs\t\t%d %d %d %d %d %d %d\n",
			procs, running, sleeping, unint,
			zombie, stopped, dead);
	}
	else
	{
		run_st = check_level_ge(running, proc_run_fail_level, -1);
		unint_st = check_level_ge(unint, proc_unint_fail_level, -1);
		printe("CTNum ");
		printc(0, "%d", venum - 1);
		if (pstat_mode) {
			printe(" VMNum ");
			printc(0, "%d", vmnum);
		}
		printe(", procs ");
		printc(0, "%4d", procs);
		printe(": R ");
		printc(run_st, "%3d", running);
		printe(", S ");
		printc(0, "%4d", sleeping);
		printe(", D ");
		printc(unint_st, "%3d", unint);
		printe(", Z ");
		printc(0, "%3d", zombie);
		printe(", T ");
		printc(0, "%3d", stopped);
		printe(", X ");
		printc(0, "%3d\n", dead);
	}
	return;
}

void show_cpu()
{
	float cpu_ves = 0, cpu_ve0 = 0;
	unsigned long sum;
	float lat_max, lat_avg;
	float sys, idle;
	int latmax_st, latavg_st, sys_st, idle_st;

	if (!get_show_status(D_CPU_INFO))
		return;
	if (total_cycles)
	{
		cpu_ves = total_cycles ? 100.0*total_time/total_cycles : 0;
		cpu_ves = CLAMP(cpu_ves, 100.0);
	}

	sum = user_j + system_j + nice_j + idle_j;
	if (!sum) sum=1;

	cpu_ve0 = 100.0*(user_j+system_j+nice_j)/sum - cpu_ves;
	cpu_ve0 = CLAMP(cpu_ve0, 100.0);
	cpu_ve0 = cpu_ve0 < 0.0 ? 0.0 : cpu_ve0;
	if (avg_mode)
		lat_max = 1000.0*sched.avg[0]/CYCLES_PER_SEC;
	else
		lat_max = 1000.0*sched.max/CYCLES_PER_SEC;
	lat_avg = 0;
	if (sched.num-sched.prev_num)
		lat_avg = 1000.0*(sched.time-sched.prev_time)/
			(sched.num-sched.prev_num)/CYCLES_PER_SEC;
	sys = 100.0*system_j/sum;
	idle = 100.0*idle_j/sum;
	latmax_st = check_level_ge(lat_max, cpu_latmax1_fail_level,
				cpu_latmax2_fail_level);
	latavg_st = check_level_ge(lat_avg, cpu_latavg1_fail_level,
				cpu_latavg2_fail_level);
	sys_st = check_level_ge(sys, cpu_sys_fail_level, -1);
	idle_st = check_level_le(idle, cpu_idle_fail_level, -1);

	if (gen_nofmt_output)
	{
		printf("CPU\t\t%.0f %.0f %.0f %.0f %.0f\n",
			cpu_ves,
			cpu_ve0,
			100.0*user_j/sum,
			sys,
			idle);
		printf("sched latency\t%.0f %.0f\n",
			lat_max,
			lat_avg);
	}
	else
	{
		printe("CPU ");
		print_status((latmax_st | latavg_st | sys_st | idle_st));
		printe(": ");
		if (venum > 1) {
			printe("CTs ");
			printc(0, "%3d", (int)cpu_ves);
			printe("%%, ");
		}
		printe("CT0 ");
		printc(0, "%3d", (int)cpu_ve0);
		printe("%%, user ");
		printc(0, "%3.0f", (float)100*user_j/sum);
		printe("%%, sys ");
		printc(sys_st, "%3.0f", sys);
		printe("%%, idle ");
		printc(idle_st, "%3.0f", idle);
		printe("%%");
		if (vz_latency_version)
		{
			printe(", lat(ms) ");
			printc(latmax_st, "%3.0f", lat_max);
			printe("/");
			printc(latavg_st, "%-.0f", lat_avg);
		}
		printe("\n");
	}
}

int get_mem_zones_status()
{
	unsigned long free,  min, low, high, active, inactive, size, i;
	int act_st1, act_st2;
	int st = 0;

	for (i=0; i<sizeof(zone_info)/sizeof(zone_info_t); i++)
	{
		if (!zone_info[i].name) break;
		sscanf(zone_info[i].name, "%*s free %lu min %lu low %lu "
			"high %lu active %lu inactive %lu size %lu",
			&free, &min, &low, &high,
			&active, &inactive, &size);

		/* convert to MB */
		free = free/(1024*1024/page_size);
		active = active/(1024*1024/page_size);
		inactive = inactive/(1024*1024/page_size);
		size = size/(1024*1024/page_size);

		act_st1 = check_level_le(active + inactive + free,
				zone_act1_fail_level*size/100,
				zone_act2_fail_level*size/100);
		act_st2 = check_level_le(active + inactive + free,
				zone_act_abs1_fail_level,
				zone_act_abs2_fail_level);
		st |= act_st1 | act_st2;
	}
	return st;
}

void show_mem()
{
	float max_lat, avg_lat;
	int latmax_st, latavg_st;

	if (!get_show_status(D_MEM_INFO))
		return;
	if (avg_mode)
	{
		max_lat = max(max(k0.avg[0], u0.avg[0]),
				max(k1.avg[0], u1.avg[0]));
		max_lat = max(max_lat, a0.avg[0]);
		avg_lat = (a0.avg[0] + k0.avg[0] + u0.avg[0] +
				k1.avg[0] + u1.avg[0])/5.0;
	}
	else
	{
		max_lat = max(max(k0.max, u0.max), max(k1.max, u1.max));
		max_lat = max(max_lat, a0.max);
		avg_lat = (a0.max + k0.max + u0.max + k1.max + u1.max)/5.0;
	}

	max_lat = 1000.0*max_lat/CYCLES_PER_SEC;
	avg_lat = 1000.0*avg_lat/CYCLES_PER_SEC;
	latmax_st = check_level_ge(max_lat, mem_latmax1_fail_level,
			mem_latmax2_fail_level);
	latavg_st = check_level_ge(avg_lat, mem_latavg1_fail_level,
			mem_latavg2_fail_level);
	if (gen_nofmt_output)
	{
		printf("Mem\t\t%lu %lu %lu\n",
			meminfo.mem_total/1024,
			meminfo.low_free/1024,
			meminfo.high_free/1024);
		if (vz_mmperf_version)
			printf("Mem latency\t%0.f %0.f\n", max_lat, avg_lat);
	}
	else
	{
		printe("Mem ");
		print_status((latmax_st | latavg_st | get_mem_zones_status()));
		printe(": total ");
		printc(0, "%lu", meminfo.mem_total/1024);
		printe("MB, free ");
		printc(0, "%lu", meminfo.mem_free/1024);
		printe("MB");
		if (vz_mmperf_version)
		{
			printe(", lat(ms) ");
			printc(latmax_st, "%.0f", max_lat);
			printe("/");
			printc(latavg_st, "%.0f", avg_lat);
		}
		printe("\n");
	}
}

void show_mem_zones()
{
	char name[26];
	unsigned long free, min, low, high, active, inactive, size, i;
	int pages_sum, nr_pages, order;
	int pages_st, act_st, act_st1, act_st2;
	char *p, *ep;

	if (!get_show_status(D_MEM_ZONES))
		return;
	for (i=0; i<sizeof(zone_info)/sizeof(zone_info_t); i++)
	{
		if (!zone_info[i].name) break;
		sscanf(zone_info[i].name, "%25s free %lu min %lu low %lu "
			"high %lu active %lu inactive %lu size %lu",
			name, &free, &min, &low, &high,
			&active, &inactive, &size);

		if (avg_mode)
		{
			active = zone_info[i].active_avg[0];
			inactive = zone_info[i].inactive_avg[0];
			free = zone_info[i].free_avg[0];
		}

		/* convert to MB */
		free = free/(1024*1024/page_size);
		min = min/(1024*1024/page_size);
		low = low/(1024*1024/page_size);
		high = high/(1024*1024/page_size);
		active = active/(1024*1024/page_size);
		inactive = inactive/(1024*1024/page_size);
		size = size/(1024*1024/page_size);

		printe("  ZONE%d (%s): size ", i, name);
		printc(0, "%lu", size);

		if(strstr(name,"DMA") != NULL)
		{
			act_st1 = C_NORMAL;
			act_st2 = C_NORMAL;
		} else
		{
			act_st1 = check_level_le(active + inactive + free,
					zone_act1_fail_level*size/100,
					zone_act2_fail_level*size/100);
			act_st2 = check_level_le(active + inactive + free,
					zone_act_abs1_fail_level,
					zone_act_abs2_fail_level);
		}
		act_st =  max(act_st1, act_st2);
		printe("MB, act ");
		printc(act_st, "%lu", active);
		printe("MB, inact ");
		printc(act_st, "%lu", inactive);
		printe("MB, free ");
		printc(0, "%lu", free);
		printe("MB (");
		printc(0, "%lu/%lu/%lu", min , low, high);
		printe(")\n");
		if (!get_show_status(D_MEM_ZONES_F))
			continue;
		/* Display fragmentatiom */
		p = zone_info[i].fragmentation;
		ep = p + strlen(p);
		pages_sum = 0;
		/* Calculate active+inactive && order > 1 */
		while (p < ep) {
			if (sscanf(p, "%d*%d", &nr_pages, &order) == 2)
				if (order > 1)
					pages_sum += nr_pages*order;
			if ((p = strchr(p, ' ')) == NULL)
				break;
			p++;
		}
		pages_st = check_level_le(pages_sum, zone_pages_fail_level, -1);
		printc(pages_st, "    fragm %s\n", zone_info[i].fragmentation);
	}
}

void show_mem_latency()
{
	float mulr = 1000.0/CYCLES_PER_SEC;

	if (!get_show_status(D_MEM_LAT))
		return;
	if (!vz_mmperf_version)
		return;
	printe("  Mem lat (ms): A0 ");
	printc(0, "%.0f", a0.max*mulr);
	printe(", K0 ");
	printc(0, "%.0f", k0.max*mulr);
	printe(", U0 ");
	printc(0, "%.0f", u0.max*mulr);
	printe(", K1 ");
	printc(0, "%.0f", k1.max*mulr);
	printe(", U1 ");
	printc(0, "%.0f\n", u1.max*mulr);
}

void show_slabinfo()
{
	if (!get_show_status(D_SLAB_INFO))
		return;
	printe("  Slab pages: ");
	printc(0, "%lu", slab_total >> MB_SHIFT);
	printe("MB/");
	printc(0, "%lu", slab_active >> MB_SHIFT);
	printe("MB (ino ");
	printc(0, "%lu", slab_inode >> MB_SHIFT);
	printe("MB, de ");
	printc(0, "%lu", slab_dentry >> MB_SHIFT);
	printe("MB, bh ");
	printc(0, "%lu", slab_bh >> MB_SHIFT);
	printe("MB, pb ");
	printc(0, "%lu", slab_pb >> MB_SHIFT);
	printe("MB)\n");
}

void show_swap()
{
	int free_st = 0, in_st = 0, out_st = 0;

	if (!get_show_status(D_SWAP_INFO))
		return;
	if (loop_num && meminfo.swap_total)
	{
		free_st = check_level_le((100ULL*meminfo.swap_free)/
				meminfo.swap_total,
				swap_free1_fail_level,
				swap_free2_fail_level);
		in_st = check_level_ge(swap_in,
				swap_in1_fail_level,
				swap_in2_fail_level);
		out_st = check_level_ge(swap_out,
				swap_out1_fail_level,
				swap_out2_fail_level);
	}
	if (gen_nofmt_output)
	{
		printf("Swap\t\t%lu %lu %.3f %.3f\n",
			meminfo.swap_total/1024,
			meminfo.swap_free/1024,
			swap_in,
			swap_out);
	}
	else
	{
		printe("Swap ");
		print_status((free_st | out_st | in_st));
		printe(": tot ");
		printc(0, "%lu", meminfo.swap_total/1024);
		printe("MB, free ");
		printc(free_st, "%lu", meminfo.swap_free/1024);
		printe("MB, in ");
		printc(in_st, "%.3f", loop_num ? swap_in : 0);
		printe("MB/s, out ");
		printc(out_st, "%.3f", loop_num ? swap_out : 0);
		printe("MB/s\n");
	}
}

void show_swap_lat()
{
	float max, avg, max_phys, avg_phys;
	int swap_st;

	if (!get_show_status(D_SWAP_LAT))
		return;
	if (avg_mode)
		max = 1000.0*swapin.avg[0]/CYCLES_PER_SEC;
	else
		max = 1000.0*swapin.max/CYCLES_PER_SEC;
	avg = 1000.0*(swapin.time-swapin.prev_time)/CYCLES_PER_SEC/
		(swapin.num-swapin.prev_num ? swapin.num-swapin.prev_num : 1);

	if (gen_nofmt_output)
	{
		printf("Swapin latency\t%lu %0.f %0.f\n",
			swapin.num,
			max,
			avg);
	}
	else
	{
		printe("  Swap lat: si ");
		printc(0, "%lu", swapin.num);
		printe(", ");
		if (vz_latency_version)
		{
			swap_st = check_level_ge(max,
				swap_lat_max1_fail_level,
				swap_lat_max2_fail_level);
			printc(swap_st, "%.0f", max);
			printe("/");
			swap_st = check_level_ge(avg,
				swap_lat_avg1_fail_level,
				swap_lat_avg2_fail_level);
			printc(swap_st, "%.0f", avg);
			printe(" ms, ");
		}
		printe("so ");
	}
	max = 1000.0*swapout.max/CYCLES_PER_SEC;
	avg = 1000.0*swapout.total/CYCLES_PER_SEC/
		(swapout.num ? swapout.num : 1);
	max_phys = 1000.0*swapout.max_phys/CYCLES_PER_SEC;
	avg_phys = 1000.0*swapout.total_phys/CYCLES_PER_SEC/
		(swapout.num ? swapout.num : 1);
	if (gen_nofmt_output)
	{
		printf("Swapout stats\t%lu %0.f %0.f %0.f %0.f\n",
			swapout.num,
			max,
			avg,
			max_phys,
			avg_phys);
	}
	else
	{
		printc(0, "%lu", swapout.num);
		if (vz_latency_version)
		{
			printe(", ");
			swap_st = check_level_ge(max,
				swap_lat_max1_fail_level,
				swap_lat_max2_fail_level);
			printc(swap_st, "%.0f", max);
			printe("/");
			swap_st = check_level_ge(avg,
				swap_lat_avg1_fail_level,
				swap_lat_avg2_fail_level);
			printc(swap_st, "%.0f", avg);
			printe(" ms, ");
			swap_st = check_level_ge(max_phys,
				swap_lat_max1_fail_level,
				swap_lat_max2_fail_level);
			printc(swap_st, "%.0f", max_phys);
			printe("/");
			swap_st = check_level_ge(avg_phys,
				swap_lat_avg1_fail_level,
				swap_lat_avg2_fail_level);
			printc(swap_st, "%.0f", avg_phys);
			printe(" cpu ms");
		}
		printe("\n");
	}
}

void show_swap_cache()
{
	unsigned long add, del, find1, find2;
	int res;

	if (!get_show_status(D_SWAPC_INFO))
		return;

	res = sscanf(swap_cache_str, "add %lu, del %lu, find %lu/%lu",
		&add,
		&del,
		&find1,
		&find2);
	if (res != 4)
		return;

	if (gen_nofmt_output)
	{
		printf("Swap cache\t%lu %lu %lu %lu\n",
			add,
			del,
			find1,
			find2);
	}
	else
	{
		printe("  Swap cache: add ");
		printc(0, "%lu", add);
		printe(", del ");
		printc(0, "%lu", del);
		printe(", find ");
		printc(0, "%lu", find1);
		printe("/");
		printc(0, "%lu", find2);
		printe("\n");
	}
}

void sort_ves()
{
	if (sort_mode < 0)
		return;
	field_ar[sort_mode].sort_fn();
}

int check_ve_filter(const char *id)
{
	int i;

	if (!veids_specified)
		return 1;
	for (i = 0; i < veids_specified; i++)
		if (strcmp(veids[i], id) == 0)
			return 1;
	return 0;
}

void show_info()
{
	attron(COLOR_PAIR(C_INVERT));
	attron(A_BOLD);
	printw("%s", info_str);
	attroff(A_BOLD);
	attroff(COLOR_PAIR(C_INVERT));
}

void _printe(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (gen_text_output)
		vprintf(fmt, ap);
	else
		vw_printw(stdscr, fmt, ap);
	va_end(ap);
}

#define F(x)	((x) >= 9.95 ? 0 : 1), (x)
#define F2(x)	((x) >= 9.95 ? 1 : 2), (x)
#define KBYTE	(1024)
#define MBYTE	(1024 * KBYTE)
#define GBYTE	(1024 * MBYTE)

#define PRINT_SCALE(scale_to, bytes)					\
	if (scale_to == SCALE_BYTES)						\
		snprintf(buf, size, "%*.*f", fw - 5, 0, (bytes / 1.));		\
	else if (scale_to == SCALE_KBYTES)					\
		snprintf(buf, size, "%*.*fKB", fw - 5, F(bytes / (1. * KBYTE)));	\
	else if (scale_to == SCALE_MBYTES)					\
		snprintf(buf, size, "%*.*fMB", fw - 5, F(bytes / (1. * MBYTE)));	\
	else if (scale_to == SCALE_GBYTES)					\
		snprintf(buf, size, "%*.*fGB", fw - 5, F(bytes / (1. * GBYTE)));

/*
 * Print total value in MBytes (GBytes)
 */
#define PRINT_TOTAL_MB(field)				\
{							\
	int scale_to;					\
	if (scale != SCALE_NONE)			\
		scale_to = scale;			\
	else if (vestat->field >= GBYTE)		\
		scale_to = SCALE_GBYTES;		\
	else if (vestat->field >= MBYTE)		\
		scale_to = SCALE_MBYTES;		\
	else						\
		scale_to = SCALE_KBYTES;		\
	PRINT_SCALE(scale_to, vestat->field)		\
} while(0);

/*
 * Print rate in Megabytes per seconds
 */
#define PRINT_RATE_MB(bytes)				\
{							\
	int scale_to;					\
	if (scale != SCALE_NONE)			\
		scale_to = scale;			\
	else if (bytes >= MBYTE)			\
		scale_to = SCALE_MBYTES;		\
	else						\
		scale_to = SCALE_KBYTES;		\
	PRINT_SCALE(scale_to, bytes)			\
} while(0);

#define PRINT_SCALE_RATE_LIMIT(scale_to, bytes, limit_bytes)		\
	if (!limit_bytes) {						\
		if (scale_to == SCALE_KBYTES)				\
			snprintf(buf, size, "%*.*f/---KB/s", fw - 8,	\
					F(bytes / (1. * KBYTE)));	\
		else if (scale_to == SCALE_MBYTES)			\
			snprintf(buf, size, "%*.*f/---MB/s", fw - 8,	\
					F(bytes / (1. * MBYTE)));	\
		else if (scale_to == SCALE_GBYTES)			\
			snprintf(buf, size, "%*.*f/---GB/s", fw - 8,	\
					F(bytes / (1. * GBYTE)));	\
		else							\
			snprintf(buf, size, "%*.*f/---B/s", fw - 8,	\
					0, (bytes / 1.));			\
	} else {							\
		if (scale_to == SCALE_KBYTES)				\
			snprintf(buf, size, "%*.*f/%*.*fKB/s",		\
				(fw - 5)/2, F(bytes / (1. * KBYTE)),	\
				(fw - 5)/2, F(limit_bytes / (1. * KBYTE)));	\
		else if (scale_to == SCALE_MBYTES)			\
			snprintf(buf, size, "%*.*f/%*.*fMB/s",		\
				(fw - 5)/2, F(bytes / (1. * MBYTE)),	\
				(fw - 5)/2, F(limit_bytes / (1. * MBYTE)));	\
		else if (scale_to == SCALE_GBYTES)			\
			snprintf(buf, size, "%*.*f/%*.*fGB/s",		\
				(fw - 5)/2, F(bytes / (1. * GBYTE)),	\
				(fw - 5)/2, F(limit_bytes / (1. * GBYTE)));	\
		else							\
			snprintf(buf, size, "%*.*f/%*.*fB/s",		\
				(fw - 5)/2, 0, (bytes / 1.),		\
				(fw - 5)/2, 0, (limit_bytes / 1.));	\
	}

/*
 * Print rate and limit in Megabytes per seconds
 */
#define PRINT_RATE_LIMIT_MB(bytes, limit_bytes)		\
{							\
	int scale_to;					\
	if (scale != SCALE_NONE)			\
		scale_to = scale;			\
	else if (!limit_bytes && bytes >= MBYTE)	\
		scale_to = SCALE_MBYTES;		\
	else if (limit_bytes >= MBYTE)			\
		scale_to = SCALE_MBYTES;		\
	else						\
		scale_to = SCALE_KBYTES;		\
	PRINT_SCALE_RATE_LIMIT(scale_to, bytes, limit_bytes)	\
} while(0)

/*
 * Print rate in items per seconds
 */
#define PRINT_RATE(field)					\
	snprintf(buf, size, "%*.*f/s", fw - 3, F(vestat->field));

#define PRINT_RATE_LIMIT(field, limit)						\
{										\
	if (!limit)								\
		snprintf(buf, size, "%*.*f/--/s", fw - 7, F(vestat->field));	\
	else									\
		snprintf(buf, size, "%*.*f/%*.*f/s",				\
						(fw - 5)/2, F(vestat->field),	\
						(fw - 5)/2, F(limit / 1.));	\
} while(0)


#define PRINT_LIMIT_COLOR(div, level1, level2)				\
	if (div >= (float)level1)					\
	{								\
		if (div >= (float)level2)				\
			printc(C_RED, field_ar[id].format, fw, buf);	\
		else							\
			printc(C_YELLOW, field_ar[id].format, fw, buf);	\
		return;							\
	}



#define PRINT_RATE_LIMIT_COLOR(value, limit, level1, level2)			\
{										\
	if (limit)								\
	{									\
		float div = 100 * (float)value / (float)limit;			\
		PRINT_LIMIT_COLOR(div, level1, level2);				\
	}									\
} while(0)


/*
 * Print one field in a column.
 * vestat - data base;
 * id     - number of key;
 * ws     - insert white space before field.
 */
static void print_field(s_line_t *vestat, unsigned id, int ws)
{
	extern int avg_mode;
	int fw;		/* field width */
	char buf[40];	/* buffer for one field */
	int space;	/* num of bytes left in string */
	int size;	/* maximum number of bytes that can be written */
	int x;
	float held_f;
	unsigned long held_l, barrier_l;

	/* Bug ? */
	if (id >= AR_SIZE(field_ar))
		return;

	memset(buf, 0, sizeof(buf));
	size = sizeof(buf) - 1;

	x = getcurx(stdscr);
	space = screen_cols - x;

	/* Insert white space */
	if (ws && space > 0) {
		printe(" ");
		space--;
	}

	/*
	 * getyx will return -1 for text mode, so there will be
	 * always the space for all fields.
	 *
	 * Do nothing if there is no space left in the string.
	 */
	if (space <= 0)
		return;

	/* Field width */
	fw = field_ar[id].width;
	if (fw > space)
		fw = space;

	if (size > space)
		size = space;

	switch (id) {
	case ID_ID:
		snprintf(buf, size, gen_text_output ? "%s" : "%8.8s",
				vestat->ve_uuid);
		break;
	case ID_ST:
		if (vestat->fcnt_st > C_YELLOW || vestat->maxlat_st > C_YELLOW){
			snprintf(buf, size, "%s", "!!");
			printc(C_RED, field_ar[id].format, fw, buf);
		} else {
			snprintf(buf, size, "%s", "OK");
			printc(C_NORMAL, field_ar[id].format, fw, buf);
		}
		return;
	case ID_VM:
		held_f = 100.0 * vestat->privvmpages.held * page_size /
                                        (meminfo.mem_total * 1024.0);
		if (vestat->mem_total)
			snprintf(buf, size, "%3.*f/%-3.*f",
				F(100.0 * (vestat->mem_total - vestat->mem_free)
					/ meminfo.mem_total),
				F(100.0 * vestat->mem_total / meminfo.mem_total));
		else if (vestat->slm)
			snprintf(buf, size, "%3.*f/%-3.*f", F(held_f),
				F(100.0 * vestat->slm_info.mem.inst_usage /
					vestat->slm_info.oomlimit));
		else if (vestat->privvmpages.barrier >= LONG_MAX)
			snprintf(buf, size, "%3.*f/" STR_MAX, F(held_f));
		else
			snprintf(buf, size, "%3.*f/%-3.*f", F(held_f),
				F(100.0 * vestat->privvmpages.barrier *
					page_size /
					(meminfo.mem_total * 1024.0)));
		break;
	case ID_SW:
		if( vestat->swap_total == 0 )
		{
			snprintf(buf, size, "0/0" );
			break;
		}
		held_f = 100.0 * (vestat->swap_total - vestat->swap_free)
			/ vestat->swap_total;
		if ( vestat->swap_total > GBYTE )
			snprintf(buf, size, "%*.*f/%*.*fTB",
				(fw - 3)/2, F(held_f),
				(fw - 3)/2,
				F(vestat->swap_total / (1. * GBYTE)));
		else if ( vestat->swap_total > MBYTE )
			snprintf(buf, size, "%*.*f/%*.*fGB",
				(fw - 3)/2, F(held_f),
				(fw - 3)/2,
				F(vestat->swap_total / (1. * MBYTE)));
		else if ( vestat->swap_total > KBYTE )
			snprintf(buf, size, "%*.*f/%*.*fMB",
				(fw - 3)/2, F(held_f),
				(fw - 3)/2,
				F(vestat->swap_total / (1. * KBYTE)));
		else if ( vestat->swap_total )
			snprintf(buf, size, "%*.*f/%*.*fKB",
				(fw - 3)/2, F(held_f),
				(fw - 3)/2,
				F(vestat->swap_total/1.));
		break;
	case ID_KM:
		held_f = 100.0 * vestat->kmemsize.held /
                                        meminfo.low_total / 1024;
		if (vestat->slm)
			snprintf(buf, size, "%3.*f/" STR_SLM_LIMIT, F(held_f));
		else if (vestat->kmemsize.barrier >= LONG_MAX)
			snprintf(buf, size, "%3.*f/" STR_MAX, F(held_f));
		else
			snprintf(buf, size, "%3.*f/%-3.*f", F(held_f),
				F(100.0 * vestat->kmemsize.barrier /
					meminfo.low_total / 1024));
		break;
	case ID_PR:
		if (vestat->numproc.barrier >= LONG_MAX) {
			snprintf(buf, size, "%lu/%lu/" STR_MAX,
				vestat->running,
				vestat->tasks);
		} else {
			snprintf(buf, size, "%lu/%lu/%lu",
				vestat->running,
				vestat->tasks,
				vestat->numproc.barrier);
		}
		break;
	case ID_CPU:
		snprintf(buf, size, "%4.*f/%-3.*f",
			F2((float)vestat->cpu_prcntg),
			F((float)vestat->cpuunits_prcntg));
		break;
	case ID_SOCK:
		held_l = vestat->numtcpsock.held +
				vestat->numothersock.held;
		barrier_l = vestat->numtcpsock.barrier +
                                vestat->numothersock.barrier;
		if (barrier_l >= LONG_MAX)
			snprintf(buf, size, "%lu/" STR_MAX, held_l);
		else
			snprintf(buf, size, "%lu/%lu", held_l, barrier_l);
		break;
	case ID_FCNT:
		if (avg_mode) {
			snprintf(buf, size, "%ld", vestat->fcnt);
			break;
		} else {
			snprintf(buf, size, "%ld", vestat->fcnt_delta);
			printc(vestat->fcnt_st, field_ar[id].format, fw, buf);
			return;
		}
	case ID_MLAT:
		snprintf(buf, size, "%.0f", (float)vestat->maxlat
						/ CYCLES_PER_SEC * 1000.0);
		printc(vestat->maxlat_st, field_ar[id].format, fw, buf);
		return;
	case ID_IP:
		if (hostname_mode && vestat->hostname != NULL)
			snprintf(buf, size, "%-*s", fw, vestat->hostname);
		else
			snprintf(buf, size, "%-*s", fw, vestat->ip_addr);
		break;
	case ID_UUID:
		if (vestat->uuid != NULL)
			snprintf(buf, size, "%-*s", fw, vestat->uuid);
		else
			snprintf(buf, size, "%ld", vestat->veid);
		break;
	case ID_NAME:
		if (name_mode && vestat->name != NULL) {
			snprintf(buf, size, "%-*s", fw, vestat->name);
		} else {
			if (vestat->uuid != NULL)
				snprintf(buf, size, "%-*s", fw, vestat->uuid);
			else
				snprintf(buf, size, "%ld", vestat->veid);
		}
		break;
	case ID_HOST:
		if (vestat->hostname != NULL)
			snprintf(buf, size, "%-*s", fw, vestat->hostname);
		else
			snprintf(buf, size, "%-*s", fw, "");
		break;
	case ID_IOR:
		PRINT_RATE_MB(vestat->io_rate[IO_READ]);
		break;
	case ID_IOW:
		PRINT_RATE_MB(vestat->io_rate[IO_WRITE]);
		break;
	case ID_IOD:
		PRINT_RATE_MB(vestat->io_rate[IO_DIRTY]);
		break;
	case ID_IOS:
		PRINT_RATE_MB(vestat->io_rate[IO_SYNC]);
		break;
	case ID_IORT:
		PRINT_TOTAL_MB(io[IO_READ]);
		break;
	case ID_IOWT:
		PRINT_TOTAL_MB(io[IO_WRITE]);
		break;
	case ID_RX:
		PRINT_RATE_MB(vestat->net.rate[NET_RX]);
		break;
	case ID_TX:
		PRINT_RATE_MB(vestat->net.rate[NET_TX]);
		break;
 	case ID_RXP:
		PRINT_RATE(net.rate[NET_RX_PKTS]);
		break;
 	case ID_TXP:
		PRINT_RATE(net.rate[NET_TX_PKTS]);
		break;
	case ID_RXT:
		PRINT_TOTAL_MB(net.total[NET_RX]);
		break;
	case ID_TXT:
		PRINT_TOTAL_MB(net.total[NET_TX]);
		break;
 	case ID_RXPT:
		snprintf(buf, size, "%*lld", fw,
			vestat->net.total[NET_RX_PKTS]);
		break;
 	case ID_TXPT:
		snprintf(buf, size, "%*lld", fw,
			vestat->net.total[NET_TX_PKTS]);
		break;
	case ID_IO_USED:
		snprintf(buf, size, "%.2f",
			vestat->io_rate[IO_USED_TM] / 10);
		break;
	case ID_IO_WAIT:
		snprintf(buf, size, "%.2f",
			vestat->io_rate[IO_WAIT_TM] / 10);
		break;
	case ID_IO_AVG_SIZE:
		snprintf(buf, size, "%*.0f", fw,
			(vestat->io_rate[IO_TRANSFERRED_SEC] /
				vestat->io_rate[IO_REQ_COMPLETED] * 512));
		break;
	case ID_IO:
		{
			float io = vestat->io_rate[IO_WRITE]+vestat->io_rate[IO_READ];
			PRINT_RATE_LIMIT_MB(io, vestat->io_limit);
			PRINT_RATE_LIMIT_COLOR(io, vestat->io_limit, disk_load1_level, disk_load2_level);
		}
		break;

	case ID_IOPS:
		PRINT_RATE_LIMIT(io_rate[IO_REQ_COMPLETED], vestat->iops_limit);
		PRINT_RATE_LIMIT_COLOR(vestat->io_rate[IO_REQ_COMPLETED], vestat->iops_limit,
					disk_load1_level, disk_load2_level);
		break;

	case ID_SWAPIN:
		PRINT_RATE(io_rate[IO_SWAPIN_COUNT]);
		break;

	case ID_CPU_IOWAIT:
		snprintf(buf, size, "%4.*f%%",
			F2((float)vestat->cpu_rate[CPU_IOWAIT]));

		PRINT_LIMIT_COLOR(vestat->cpu_rate[CPU_IOWAIT],
					disk_load2_level, disk_load2_level);
		break;
	case ID_CPU_IOWAIT_AVG:
		snprintf(buf, size, "%4.*f%%",
			F2((float)vestat->cpu_rate_avg[CPU_IOWAIT]));

		PRINT_LIMIT_COLOR(vestat->cpu_rate_avg[CPU_IOWAIT],
					disk_load2_level, disk_load2_level);
		break;
	default:
		/* What is it? */
		printe("Illegal field!");
		return;
	}

	printe(field_ar[id].format, fw, buf);
}

/*
 * Show table caption. It depends on selected keys format.
 */
void show_caption(void)
{
	int fid;
	unsigned f;
	char buf[MAX_CHARS + 1];
	char *p = buf;
	int n;
	int space;
	int found = 0;
	int pad;

	space = screen_cols + 1;

	/*
	 * Draw table head through the whole screen (like in 'top')
	 */
	if (gen_text_output)
		pad = 0;
	else
		pad = ' ';

	memset(buf, pad, screen_cols);
	buf[screen_cols] = 0;

	for (f = 0; f < FIELDS_COUNT; f++) {
		fid = field_set[f];
		if (!fid)
			continue;
		if (found++) {
			/* Insert mandatory column delimiter before */
			*p++ = ' ';
			space--;
		}
		/*
		 * Backward compatibility...
		 */
		if (fid == ID_IP && hostname_mode)
			n = snprintf(p, space, field_ar[fid].format,
				field_ar[fid].width, "HOSTNAME");
		else
			n = snprintf(p, space, field_ar[fid].format,
				field_ar[fid].width, field_ar[fid].head);
		/* Overflow ? */
		if (n >= space)
			n = space;
		if (n >= 0) {
			p += n;
			space -= n;
		}
		/* Avoid overflow */
		if (space == 0)
			break;
	}
	/* Replace '\0' by ' ' */
	*p++ = ' ';

	if (gen_text_output) {
		_printe(buf);
	} else {
		attron(COLOR_PAIR(C_INVERT));
		attron(A_BOLD);
		_printe(buf);
		attroff(A_BOLD);
		attroff(COLOR_PAIR(C_INVERT));
	}
}

void show_ves(void)
{
	int idx = 0;
	int fid, y;
	unsigned f;
	int found;

	y = getcury(stdscr);

	/*
	 * We can exceed rows count before showing the caption.
	 */
	if (y >= screen_rows)
		return;

	show_caption();
	if (gen_text_output)
		printe("\n");
	else
		move (++y, 0);

	sort_ves();
	while (1)
	{
		s_line_t *vestat;

		if (y >= screen_rows)
			break;

		vestat = vzstatcpu_get_ve_stats(idx++, 2);
		if (!vestat)
			break;
		// Skip Vm in vzstat mode
		if (!pstat_mode && vestat->env_type != CT_TYPE)
			continue;
		// FIXME: surpress stopped Vm
		if (vestat->numproc.held == 0)
			continue;
		if (!check_ve_filter(vestat->ve_uuid))
			continue;
		if (!idle_ve_display && vestat->cycles &&
			!(vestat->used_time*100/vestat->cycles))
			continue;

		if (vestat->hostname == NULL)
			parse_ve_conf(vestat);

		found = 0;
		for (f = 0; f < FIELDS_COUNT; f++) {
			int newy;

			fid = field_set[f];
			if (!fid)
				continue;
			print_field(vestat, fid, found++);
			newy = getcury(stdscr);
			/*
			 * ncurses inserts carridge return after last
			 * symbol was printed in a window.
			 */
			if (newy != y)
				break;
		}
		if (gen_text_output)
			printe("\n");
		else
			move (++y, 0);
	}
}

void show_netstats()
{
	float rx, tx, pt, pr;
	unsigned long i;

	if (!get_show_status(D_NET_INFO))
		return;
	rx = tx = pt = pr = 0;
	if (loop_num)
		for (i=0; i<ifcount; i++)
		{
			rx += ifstats[i].rx;
			tx += ifstats[i].tx;
			pr += ifstats[i].pr;
			pt += ifstats[i].pt;
		}
	if (gen_nofmt_output)
	{
		printf("Net stats\t%.3f %0.f %.3f %0.f\n",
			rx/1024.0/1024.0,
			pr,
			tx/1024.0/1024.0,
			pt);
	}
	else
	{
		printe("Net ");
		print_status(0);
		printe(": tot: in ");
		printc(0, "%6.3f", rx/1024.0/1024.0);
		printe("MB/s ");
		printc(0, "%4.0f", pr);
		printe("pkt/s, out ");
		printc(0, "%6.3f", tx/1024.0/1024.0);
		printe("MB/s ");
		printc(0, "%4.0f", pt);
		printe("pkt/s\n");
	}
	if (!get_show_status(D_NET_IFACE))
		return;
	/* per interface statistics */
	for (i=0; i<ifcount; i++)
	{
		if (!get_show_status(D_NET_VETH) &&
				(!strncmp(ifstats[i].name, "veth", 4) ||
				!strncmp(ifstats[i].name, "vme", 3)))
			continue;
		if (loop_num)
		{
			rx = ifstats[i].rx;
			tx = ifstats[i].tx;
			pr = ifstats[i].pr;
			pt = ifstats[i].pt;
		}
		if (gen_nofmt_output)
		{
			printf("if %s stats\t%.3f %0.f %.3f %0.f\n",
				ifstats[i].name,
				rx/1024.0/1024.0,
				pr,
				tx/1024.0/1024.0,
				pt);
		}
		else
		{
			printe("%15s: in ", ifstats[i].name);
			printc(0, "%6.3f", rx/1024/1024);
			printe("MB/s ");
			printc(0, "%4.0f", pr);
			printe("pkt/s, out ");
			printc(0, "%6.3f", tx/1024/1024);
			printe("MB/s ");
			printc(0, "%4.0f", pt);
			printe("pkt/s\n");
		}
	}
}

int get_disk_status()
{
	int i, ds_st, dino_st, dspace, dinodes;
	int st = 0;

	for (i = 0; i < MAX_DISK && diskinfo[i].mnt != NULL; i++)
	{
		if (!diskinfo[i].inuse)
			continue;
		dspace = (float)diskinfo[i].disk_free / diskinfo[i].disk_total
			* 100;
		dinodes = (float)diskinfo[i].inodes_free /
			(diskinfo[i].inodes_total ? diskinfo[i].inodes_total:1)
		       	* 100;
		ds_st = check_level_le(dspace, disk_free_space1_fail_level,
				disk_free_space2_fail_level);
		if (diskinfo[i].inodes_total == 0)
			dino_st = 0;
		else
			dino_st = check_level_le(dinodes, disk_free_inodes1_fail_level,
				disk_free_inodes2_fail_level);
		st |= ds_st | dino_st;
	}
	return st;
}

void show_disks()
{
	int i;
	int dspace, dinodes;
	int ds_st, dino_st;
	float w_sum = 0, r_sum = 0;

	if (!get_show_status(D_DISK_INFO))
		return;
	/* Calculate sum r/w  by all dev */
	for (i = 0; i < MAX_DISK; i++)
	{
		if (!diskstat[i].major && !diskstat[i].minor)
			break;
		/* block is a sector 512 bytes size */
		r_sum += diskstat[i].delta_rblk/2/1024.0;
		w_sum += diskstat[i].delta_wblk/2/1024.0;
	}

	if (gen_nofmt_output)
	{
		printf("Disks stats\t%.3f %.3f\n",
			r_sum,
			w_sum);
	}
	else
	{
		printe("Disks ");
		print_status(get_disk_status());
		printe(": in ");
		printc(0, "%.3f", r_sum);
		printe("MB/s, out ");
		printc(0, "%.3f", w_sum);
		printe("MB/s\n");
	}

	if (!get_show_status(D_DISK_MNT))
		return;
	for (i = 0; i < MAX_DISK && diskinfo[i].mnt != NULL; i++)
	{
		if (!diskinfo[i].inuse)
			continue;
		dspace = (float)diskinfo[i].disk_free / diskinfo[i].disk_total
			* 100;
		dinodes = (float)diskinfo[i].inodes_free /
			(diskinfo[i].inodes_total ? diskinfo[i].inodes_total:1)
			* 100;
		ds_st = check_level_le(dspace, disk_free_space1_fail_level,
				disk_free_space2_fail_level);
		dino_st = check_level_le(dinodes, disk_free_inodes1_fail_level,
				disk_free_inodes2_fail_level);
		if (gen_nofmt_output)
		{
			printf("Disk %s(%s)\t%Lu %d %Lu %d\n",
				diskinfo[i].dev,
				diskinfo[i].mnt,
				diskinfo[i].disk_free / 1024 / 1024,
				dspace,
				diskinfo[i].inodes_free,
				diskinfo[i].inodes_total ? dinodes : 100);
		}
		else
		{
			printe("  %s(%s) free: ", diskinfo[i].dev, diskinfo[i].mnt);
			printc(0, "%lu", diskinfo[i].disk_free / 1024 / 1024);
			printe("MB");
			printc(ds_st, "(%d%%)", dspace);
			printe(", ");
			printc(0, "%lu", diskinfo[i].inodes_free);
			printe("ino");
			printc(diskinfo[i].inodes_total ? dino_st : 0, "(%d%%)\n",
					dinodes);
		}
	}
}

void show_headers()
{
	show_overall_stat();
	show_proc_stat();
	show_cpu();
	show_mem();
	show_mem_zones();
	show_mem_latency();
	show_slabinfo();
	show_swap();
	show_swap_lat();
	show_swap_cache();
	show_netstats();
	show_disks();
}

void print_text_stats()
{
	parse_base_proc_files();
	usleep(500*1000);
	loop_num++;
	parse_base_proc_files();
	show_headers();
	printe("\n");

	/* sort_mode can be already set by -s option */
	if (sort_mode < 0) {
		if (io_mode)
			sort_mode = ID_IOW;
		if (net_mode)
			sort_mode = ID_RX;
		else
			sort_mode = ID_CPU;
	}
	show_ves();
}

void clear_the_screen()
{
	if (clear_full_screen)
		clear();
	else
		erase();
}

static char *help_ar[] = {
	"Interactive commands are:\n",
	"\n",
	"h, ?\tPrint this help list\n",
	"space\tUpdate display\n",
	"t\tSpecifies the delay between screen updates\n",
	"r\tRedraw full screen from scratch\n",
	"q, Q\tQuit\n",
	"\n",
	"b\t'brief', brief output\n",
	"s\t'standart', standart output\n",
	"v\t'verbose', verbose output\n",
	"\n",
	"a\tToggles average mode\n",
	"i\tToggle display of idle CTs\n",
	"l\tToggle display of load average\n",
	"p\tToggle display of proc info\n",
	"c\tToggle display of cpu info\n",
	"m, M\tToggle display of mem info\n",
	"n, N\tToggle display of net info\n",
	"w, W\tToggle display of swap info\n",
	"d, D\tToggle display of disk info\n",
	"e\tToggle display of hostname/ip\n",
	"I\tToggle display of IO accounting\n",
	"A\tToggle display of IO statistics\n",
	"X\tToggle display of network statistics\n",
	"0\tToggle display of CT0 statistics\n",
	"\n",
	"o\tSort menu\n",
	NULL
};

void show_help()
{
	int i, n = 0;
	int maxy = getmaxy(stdscr);

	clear_full_screen = 1;
	do {
		clear_the_screen();

		for (i = 0; i < maxy - 2 && help_ar[n]; i++, n++)
			printe(help_ar[n]);
		printe("\n");
		if (help_ar[n])
			printe("Press any key to continue ...");
		else
			printe("Press any key to return to main window");

		refresh();
		timeout(-1);
		getch();
		timeout(sleep_time);
	} while (help_ar[n]);
}

void handle_delay_key()
{
	int delay;

	clear_full_screen = 1;
	clear_the_screen();
	echo();
	printe("\n\nEnter update timeout: ");
	refresh();
	timeout(-1);
	if (scanw("%d", &delay) == 1)
		if (delay >= 0)
			sleep_time = delay * 1000;
	if (sleep_time < 100) sleep_time = 100;
	timeout(sleep_time);
	noecho();
}

int handle_sort_keys()
{
	int ch, id;
	unsigned i;

	clear_full_screen = 1;
	clear_the_screen();
	printe("\n\nAvailable sort options:\n\n");

	/* Sort keys depends on key set */
	for (i = 0; i < FIELDS_COUNT; i++) {
		id = field_set[i];
		if (!id)
			break;
		do {
			if (field_ar[id].sort_fn && field_ar[id].sort_opt &&
					field_ar[id].sort_desc) {
				printe("%c\t%s\n", field_ar[id].sort_opt,
					field_ar[id].sort_desc);
				id = field_ar[id].next_id;
			} else
				id = 0;
		} while (id);
	}

	refresh();
	timeout(-1);
	ch = getch();
	sort_mode = -1;

	/* Sort keys depends on key set */
	for (i = 0; i < FIELDS_COUNT; i++) {
		id = field_set[i];
		if (!id)
			break;
		if (!field_ar[id].sort_fn || !field_ar[id].sort_opt ||
				!field_ar[id].sort_desc)
			continue;
		do {
			if (field_ar[id].sort_opt == ch) {
				sort_mode = id;
				break;
			} else {
				id = field_ar[id].next_id;
			}
		} while (id);
	}

	timeout(sleep_time);
	return 0;
}

int handle_keys()
{
	int ch;

	timeout(loop_num ? sleep_time : 100);
	ch = getch();

	strcpy(info_str, "\n");
	switch (ch)
	{
	case '?':
	case 'h':
		show_help();
		break;
	case 'q':
	case 'Q':
		return -1;
	case 'v':
		disp_flags = VL_VERBOSE;
		break;
	case 'b':
		disp_flags = VL_BRIEF;
		break;
	case 's':
		disp_flags = VL_NORMAL;
		break;
	case 'l':
		disp_flags ^= MAKE_MASK(D_LOADAVG);
		break;
	case 'p':
		disp_flags ^= MAKE_MASK(D_PROC_INFO);
		break;
	case 'c':
		disp_flags ^= MAKE_MASK(D_CPU_INFO);
		break;
	case 'm':
		if (BIT_IS_SET(disp_flags, D_MEM_ZONES_F))
			disp_flags &= ~MAKE_MASK(D_MEM_ZONES_F);
		else if (BIT_IS_SET(disp_flags, D_MEM_ZONES))
		{
			disp_flags &= ~MAKE_MASK(D_MEM_ZONES);
			disp_flags &= ~MAKE_MASK(D_MEM_LAT);
			disp_flags &= ~MAKE_MASK(D_SLAB_INFO);
		}
		else if (BIT_IS_SET(disp_flags, D_MEM_INFO))
			disp_flags &= ~MAKE_MASK(D_MEM_INFO);
		break;
	case 'M':
		if (!BIT_IS_SET(disp_flags, D_MEM_INFO))
			disp_flags |= MAKE_MASK(D_MEM_INFO);
		else if (!BIT_IS_SET(disp_flags, D_MEM_ZONES))
		{
			disp_flags |= MAKE_MASK(D_MEM_ZONES);
			disp_flags |= MAKE_MASK(D_MEM_LAT);
			disp_flags |= MAKE_MASK(D_SLAB_INFO);
		}
		else if (!BIT_IS_SET(disp_flags, D_MEM_ZONES_F))
			disp_flags |= MAKE_MASK(D_MEM_ZONES_F);
		break;
	case 'n':
		if (BIT_IS_SET(disp_flags, D_NET_IFACE))
			disp_flags &= ~MAKE_MASK(D_NET_IFACE);
		else if (BIT_IS_SET(disp_flags, D_NET_INFO))
			disp_flags &= ~MAKE_MASK(D_NET_INFO);
		break;
	case 'N':
		if (!BIT_IS_SET(disp_flags, D_NET_INFO))
			disp_flags |= MAKE_MASK(D_NET_INFO);
		else if (!BIT_IS_SET(disp_flags, D_NET_IFACE))
			disp_flags |= MAKE_MASK(D_NET_IFACE);
		break;
	case 'w':
		if (BIT_IS_SET(disp_flags, D_SWAPC_INFO))
			disp_flags &= ~MAKE_MASK(D_SWAPC_INFO);
		else if (BIT_IS_SET(disp_flags, D_SWAP_LAT))
			disp_flags &= ~MAKE_MASK(D_SWAP_LAT);
		else if (BIT_IS_SET(disp_flags, D_SWAP_INFO))
			disp_flags ^= MAKE_MASK(D_SWAP_INFO);
		break;
	case 'W':
		if (!BIT_IS_SET(disp_flags, D_SWAP_INFO))
			disp_flags |= MAKE_MASK(D_SWAP_INFO);
		else if (!BIT_IS_SET(disp_flags, D_SWAP_LAT))
			disp_flags |= MAKE_MASK(D_SWAP_LAT);
		else if (!BIT_IS_SET(disp_flags, D_SWAPC_INFO))
			disp_flags |= MAKE_MASK(D_SWAPC_INFO);
		break;
	case 'd':
		if (BIT_IS_SET(disp_flags, D_DISK_MNT))
			disp_flags &= ~MAKE_MASK(D_DISK_MNT);
		else if (BIT_IS_SET(disp_flags, D_DISK_INFO))
			disp_flags &= ~MAKE_MASK(D_DISK_INFO);
		break;
	case 'D':
		if (!BIT_IS_SET(disp_flags, D_DISK_INFO))
			disp_flags |= MAKE_MASK(D_DISK_INFO);
		else if (!BIT_IS_SET(disp_flags, D_DISK_MNT))
			disp_flags |= MAKE_MASK(D_DISK_MNT);
		break;
	case 'o':
		handle_sort_keys();
		break;
	case 't':
		handle_delay_key();
		break;
	case 'i':
		idle_ve_display ^= 1;
		sprintf(info_str, "Turn idle CTs %s\n",
				idle_ve_display ? "on" : "off");
		break;
	case 'a':
		avg_mode ^= 1;
		sprintf(info_str, "Turn average mode %s\n",
				avg_mode ? "on" : "off");
		break;
	case 'r':
		clear_full_screen = 1;
		sprintf(info_str, "Redraw full screen\n");
		break;
	case 'e':
		hostname_mode ^= 1;
		sprintf(info_str, "Turn %s mode on\n",
				hostname_mode ? "hostname" : "ip");
		break;
	case 'f':
		name_mode ^= 1;
		sprintf(info_str, "Turn %s mode on\n",
				name_mode ? "name" : "id");
		break;

	case 'I':
		io_mode ^= 1;
		io_stat_mode = 0;
		net_mode = 0;
		/* Restore display and sort modes */
		if (io_mode)
			field_set = get_field_set(field_set_io);
		else
			field_set = get_field_set(field_set_def);
		sprintf(info_str, "IO mode is %s\n",
				io_mode ? "on" : "off");
		break;
	case 'X':
		io_mode = 0;
		io_stat_mode = 0;
		net_mode ^= 1;
		/* Restore display and sort modes */
		if (net_mode)
			field_set = get_field_set(field_set_net);
		else
			field_set = get_field_set(field_set_def);
		sprintf(info_str, "NETWORK mode is %s\n",
				net_mode ? "on" : "off");
		break;
	case 'A':
		io_stat_mode ^= 1;
		io_mode = 0;
		net_mode = 0;
		/* Restore display and sort modes */
		if (io_stat_mode)
			field_set = get_field_set(field_set_io_stat);
		else
			field_set = get_field_set(field_set_def);
		sprintf(info_str, "IO stat mode is %s\n",
				io_stat_mode ? "on" : "off");
		break;
	case '0':
		show_ve0_mode ^= 1;
		sprintf(info_str, "Show VE0 mode is %s\n",
				show_ve0_mode ? "on" : "off");

		break;
	default:
		return 1;
	}
	return 0;
}

void mainloop()
{
	int res;

	clear_full_screen = 1;
	need_resize = 0;
	while(1)
	{
		if (need_resize) {
			set_win_size();
			need_resize = 0;
		}
		clear_the_screen();
		clear_full_screen = 0;
		parse_base_proc_files();
		show_headers();
		show_info();
		show_ves();
		refresh();

		res = handle_keys();
		if (res == -1) break;

		loop_num++;
		if (num_disp && loop_num >= num_disp)
			break;
	}
}

void set_color(int color)
{
	if (has_color)
		attron(COLOR_PAIR(color));
	return;
}

void unset_color(int color)
{
	if (has_color)
		attroff(COLOR_PAIR(color));
	return;
}

void print_status(int fail)
{
	int color = (fail & C_RED) ? C_RED :
		(fail & C_YELLOW ? C_YELLOW : C_GREEN);
	char *s = (fail & C_RED) ? FAIL_ST : OK_ST;

	if (has_color)
		set_color(color);
	_printe("%s", s);
	if (has_color)
		unset_color(color);

	return;
}

void printe(char *fmt, ...)
{
	va_list ap;
	char buf[128];
	int color = C_NORMAL;

	if (has_color)
		set_color(color);
	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	_printe("%s", buf);
	if (has_color)
		unset_color(color);
}

void printc(int color, char *fmt, ...)
{
	va_list ap;
	char buf[128];

	if(color == 0) color = C_NORMAL;

	if (has_color)
	{
		attron(A_BOLD);
		set_color(color);
	}
	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	_printe("%s", buf);
	if (has_color)
	{
		unset_color(color);
		attroff(A_BOLD);
	}
}

void init_colors()
{
	if (has_color == 0)
		return;
	if (has_colors() == FALSE)
		return;
	if (start_color())
		return;
	init_pair(C_RED, COLOR_RED, COLOR_BLACK);
	init_pair(C_GREEN, COLOR_GREEN, COLOR_BLACK);
	init_pair(C_YELLOW, COLOR_YELLOW, COLOR_BLACK);
	init_pair(C_NORMAL, COLOR_WHITE, COLOR_BLACK);
	init_pair(C_INVERT, COLOR_BLACK, COLOR_WHITE);
	has_color = 1;
	return;
}

void get_page_size()
{
	long pagesize;
	if ((pagesize = sysconf(_SC_PAGESIZE)) == -1)
		return;
	page_size = pagesize;
}

/*
 * Set the screen dimensions and arrange for the real workhorse.
 * (also) catches:
 *    SIGWINCH and SIGCONT
 */
static void set_win_size (void)
{
	char *env_columns;  // Unix98 environment variable COLUMNS
	char *env_lines;    // Unix98 environment variable LINES
	long t;
	char *endptr;

	/*
	 * Get new screen dimensions for non-text mode.
	 */
	if (!gen_text_output) {
		struct winsize wz;

		if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &wz) != -1
				&& wz.ws_col > 0 && wz.ws_row > 0) {
			screen_cols = wz.ws_col;
			screen_rows = wz.ws_row;
		}
	}

	/*
	 * User can override settings by envvar's
	 */
	env_columns = getenv("COLUMNS");
	if(env_columns && *env_columns) {
		t = strtol(env_columns, &endptr, 0);
		if(!*endptr && (t > 0) && (t <= MAXINT))
			screen_cols = (int)t;
	}
	env_lines   = getenv("LINES");
	if(env_lines && *env_lines) {
		t = strtol(env_lines, &endptr, 0);
		if(!*endptr && (t > 0) && (t <= MAXINT))
			screen_rows = (int)t;
	}
#if 0	/* Let think that in 2010 all terminal emulators are good until reported reverse */
	/*
	 * FIXME: We have to determine do we need to avoid last column.
	 * In the current version it will be skipped always.
	 */
	screen_cols--;
#else
	/*
	 * be crudely tolerant of crude tty emulators
	 */
	if (avoid_last_column)
		screen_cols--;
#endif

	/*
	 * we might disappoint some folks (but they'll deserve it)
	 */
	if (MAX_CHARS < screen_cols)
		screen_cols = MAX_CHARS;

	if (gen_text_output)
		return;

	/*
	 * change the curses terminal size
	 */
	resizeterm(screen_rows, screen_cols);
	wrefresh(stdscr);
	clear();
}

static void sighandler_resize (int dont_care_sig)
{
	(void)dont_care_sig;
	need_resize = 1;
}

static void sighandler_term (int dont_care_sig)
{
	(void)dont_care_sig;
	screen_close();
	exit(0);
}

void screen_open(void)
{
	if (gen_text_output)
		return;
	/*
	 * Save terminal state to restore it on exit
	 */
	if (tcgetattr(STDIN_FILENO, &saved_tty) == -1)
		exit_error("failed tty get");
	term_saved = 1;

	setupterm(NULL, STDOUT_FILENO, NULL);

	initscr();
	nonl();
	cbreak();
	keypad(NULL, 1);
	noecho();
	init_colors();
	wrefresh(stdscr);
	clear();

	signal(SIGQUIT,  sighandler_term);
	signal(SIGTERM,  sighandler_term);
	signal(SIGCONT,  sighandler_resize);
	signal(SIGWINCH, sighandler_resize);
}

void screen_close(void)
{
	if (gen_text_output)
		return;
	if (!term_saved)
		return;

	keypad(NULL, 0);
	nocbreak();
	endwin();
	fprintf(stdout, "\n");
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_tty);
}

static void free_diskinfo(diskinfo_t *p)
{
	free(p->dev);
	free(p->mnt);
}

static void free_ifstat(ifstat_t *p)
{
	free(p->name);
}

static void free_zone_info(zone_info_t *p)
{
	free(p->fragmentation);
	free(p->name);
}

static void cleanup_static()
{
	unsigned i;
	for (i = 0; i < AR_SIZE(diskinfo); ++i)
		free_diskinfo(&diskinfo[i]);
	for (i = 0; i < AR_SIZE(ifstats); ++i)
		free_ifstat(&ifstats[i]);
	for (i = 0; i < AR_SIZE(zone_info); ++i)
		free_zone_info(&zone_info[i]);
	free_ves();
}

int main(int argc, char **argv)
{
	if (strcmp(basename(argv[0]), "pstat") == 0)
	{
		pstat_mode = 1;
		field_set = field_set_pstat_def;
	}

	parse_options(argc, argv);
	parse_config();
	get_page_size();
	vzstatcpu_init();

	screen_open();
	set_win_size();

	if (gen_nofmt_output || gen_text_output)
		do
			print_text_stats();
		while( timeout_specified && !usleep( (sleep_time - 500) * 1000 ) );
	else
		mainloop();
	screen_close();
	close_vzctlfd();
	cleanup_static();
	return 0;
}
